from typing import Protocol

from trajax.types.predictors.common import ObstacleStatesHistory

from numtypes import Array, Dims


class NumPyObstacleStatesHistory[T: int, D_o: int, K: int, ObstacleStatesForTimeStepT](
    ObstacleStatesHistory[T, D_o, K, ObstacleStatesForTimeStepT], Protocol
):
    @property
    def array(self) -> Array[Dims[T, D_o, K]]:
        """Returns the obstacle state history as a NumPy array."""
        ...



from .registry import (
    NoCollectedDataWarning as NoCollectedDataWarning,
    CollectorRegistry as CollectorRegistry,
)
from .obstacles import (
    ObstacleStateCollector as ObstacleStateCollector,
    ObstacleForecastCollector as ObstacleForecastCollector,
)
from .risk import RiskCollector as RiskCollector
from .state import (
    StateCollector as StateCollector,
    ControlCollector as ControlCollector,
    TrajectoryCollector as TrajectoryCollector,
)
from .access import access as access
from .factory import collectors as collectors



from typing import Never, Self, overload, cast, Sequence, Final, Any
from dataclasses import dataclass

from trajax.types import (
    DataType,
    NumPyState,
    NumPyStateSequence,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPyBicycleObstacleStatesHistory,
    BicycleState,
    BicycleStateSequence,
    BicycleStateBatch,
    BicycleControlInputSequence,
    BicycleControlInputBatch,
    BicyclePositions,
    BicycleD_x,
    BICYCLE_D_X,
    BicycleD_u,
    BICYCLE_D_U,
    DynamicalModel,
    ObstacleModel,
    EstimatedObstacleStates,
)

from numtypes import Array, Dims, D, shape_of, array

import numpy as np

NO_LIMITS: Final = (float("-inf"), float("inf"))

type StateArray = Array[Dims[BicycleD_x]]
type ControlInputSequenceArray[T: int] = Array[Dims[T, BicycleD_u]]
type StateBatchArray[T: int, M: int] = Array[Dims[T, BicycleD_x, M]]
type ControlInputBatchArray[T: int, M: int] = Array[Dims[T, BicycleD_u, M]]

type StatesAtTimeStep[M: int] = Array[Dims[BicycleD_x, M]]
type ControlInputsAtTimeStep[M: int] = Array[Dims[BicycleD_u, M]]


@dataclass(frozen=True)
class NumPyBicycleState(BicycleState, NumPyState[BicycleD_x]):
    _array: StateArray

    @staticmethod
    def create(
        *, x: float, y: float, heading: float, speed: float
    ) -> "NumPyBicycleState":
        """Creates a NumPy bicycle state from individual state components."""
        return NumPyBicycleState(array([x, y, heading, speed], shape=(BICYCLE_D_X,)))

    def __array__(self, dtype: DataType | None = None) -> StateArray:
        return self.array

    @property
    def dimension(self) -> BicycleD_x:
        return self.array.shape[0]

    @property
    def x(self) -> float:
        """Returns the x position of the bicycle."""
        return self.array[0]

    @property
    def y(self) -> float:
        """Returns the y position of the bicycle."""
        return self.array[1]

    @property
    def heading(self) -> float:
        """Returns the heading (orientation) of the bicycle."""
        return self.array[2]

    @property
    def speed(self) -> float:
        """Returns the speed (velocity magnitude) of the bicycle."""
        return self.array[3]

    @property
    def array(self) -> StateArray:
        return self._array


@dataclass(kw_only=True, frozen=True)
class NumPyBicycleStateSequence[T: int, M: int = Any](
    BicycleStateSequence, NumPyStateSequence[T, BicycleD_x]
):
    batch: "NumPyBicycleStateBatch[T, M]"
    rollout: int

    @staticmethod
    def of_states[T_: int = int](
        states: Sequence[NumPyBicycleState], *, horizon: T_ | None = None
    ) -> "NumPyBicycleStateSequence[T_, D[1]]":
        """Creates a NumPy bicycle state sequence from a sequence of bicycle states."""
        assert len(states) > 0, "States sequence must not be empty."

        horizon = horizon if horizon is not None else cast(T_, len(states))
        array = np.stack([state.array for state in states], axis=0)[:, :, np.newaxis]

        assert shape_of(array, matches=(horizon, BICYCLE_D_X, 1))

        return NumPyBicycleStateSequence(
            batch=NumPyBicycleStateBatch.wrap(array), rollout=0
        )

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, BicycleD_x]]:
        return self.array

    def step(self, index: int) -> NumPyBicycleState:
        return NumPyBicycleState(self.array[index])

    def batched(self) -> "NumPyBicycleStateBatch[T, D[1]]":
        return NumPyBicycleStateBatch.wrap(self.array[..., np.newaxis])

    def x(self) -> Array[Dims[T]]:
        return self.array[:, 0]

    def y(self) -> Array[Dims[T]]:
        return self.array[:, 1]

    def heading(self) -> Array[Dims[T]]:
        return self.array[:, 2]

    def speed(self) -> Array[Dims[T]]:
        return self.array[:, 3]

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return self.array.shape[1]

    @property
    def array(self) -> Array[Dims[T, BicycleD_x]]:
        return self.batch.array[:, :, self.rollout]


@dataclass(frozen=True)
class NumPyBicycleStateBatch[T: int, M: int](
    BicycleStateBatch[T, M], NumPyStateBatch[T, BicycleD_x, M]
):
    _array: StateBatchArray[T, M]

    @staticmethod
    def wrap[T_: int, M_: int](
        array: StateBatchArray[T_, M_],
    ) -> "NumPyBicycleStateBatch[T_, M_]":
        """Creates a NumPy bicycle state batch from the given array."""
        return NumPyBicycleStateBatch(array)

    @staticmethod
    def of_states[T_: int = int](
        states: Sequence[NumPyBicycleState], *, horizon: T_ | None = None
    ) -> "NumPyBicycleStateBatch[int, D[1]]":
        """Creates a NumPy bicycle state batch from a sequence of bicycle states."""
        assert len(states) > 0, "States sequence must not be empty."

        horizon = horizon if horizon is not None else cast(T_, len(states))
        array = np.stack([state.array for state in states], axis=0)[:, :, np.newaxis]

        assert shape_of(array, matches=(horizon, BICYCLE_D_X, 1))

        return NumPyBicycleStateBatch(array)

    def __array__(self, dtype: DataType | None = None) -> StateBatchArray[T, M]:
        return self.array

    def heading(self) -> Array[Dims[T, M]]:
        return self.array[:, 2, :]

    def speed(self) -> Array[Dims[T, M]]:
        return self.array[:, 3, :]

    def rollout(self, index: int) -> NumPyBicycleStateSequence[T, M]:
        return NumPyBicycleStateSequence(batch=self, rollout=index)

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return self.array.shape[1]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[2]

    @property
    def positions(self) -> "NumPyBicyclePositions[T, M]":
        return NumPyBicyclePositions(batch=self)

    @property
    def array(self) -> StateBatchArray[T, M]:
        return self._array


@dataclass(frozen=True)
class NumPyBicyclePositions[T: int, M: int](BicyclePositions[T, M]):
    batch: NumPyBicycleStateBatch[T, M]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D[2], M]]:
        return self.batch.array[:, :2, :]

    def x(self) -> Array[Dims[T, M]]:
        return self.batch.array[:, 0, :]

    def y(self) -> Array[Dims[T, M]]:
        return self.batch.array[:, 1, :]


@dataclass(frozen=True)
class NumPyBicycleControlInputSequence[T: int](
    BicycleControlInputSequence[T], NumPyControlInputSequence[T, BicycleD_u]
):
    _array: ControlInputSequenceArray[T]

    @staticmethod
    def zeroes[T_: int](horizon: T_) -> "NumPyBicycleControlInputSequence[T_]":
        """Creates a zeroed control input sequence for the given horizon."""
        array = np.zeros((horizon, BICYCLE_D_U))

        assert shape_of(array, matches=(horizon, BICYCLE_D_U))

        return NumPyBicycleControlInputSequence(array)

    def __array__(self, dtype: DataType | None = None) -> ControlInputSequenceArray[T]:
        return self.array

    @overload
    def similar(self, *, array: Array[Dims[T, BicycleD_u]]) -> Self: ...

    @overload
    def similar[L: int](
        self, *, array: Array[Dims[L, BicycleD_u]], length: L
    ) -> "NumPyBicycleControlInputSequence[L]": ...

    def similar[L: int](
        self, *, array: Array[Dims[L, BicycleD_u]], length: L | None = None
    ) -> "Self | NumPyBicycleControlInputSequence[L]":
        # NOTE: "Wrong" cast to silence the type checker.
        effective_length = cast(T, length if length is not None else array.shape[0])

        assert shape_of(
            array, matches=(effective_length, self.dimension), name="similar array"
        )

        return self.__class__(array)

    def accelerations(self) -> Array[Dims[T]]:
        return self.array[:, 0]

    def steering_angles(self) -> Array[Dims[T]]:
        return self.array[:, 1]

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_u:
        return self.array.shape[1]

    @property
    def array(self) -> ControlInputSequenceArray[T]:
        return self._array


@dataclass(frozen=True)
class NumPyBicycleControlInputBatch[T: int, M: int](
    BicycleControlInputBatch[T, M], NumPyControlInputBatch[T, BicycleD_u, M]
):
    _array: ControlInputBatchArray[T, M]

    @staticmethod
    def zero[T_: int, M_: int](
        *, horizon: T_, rollout_count: M_ = 1
    ) -> "NumPyBicycleControlInputBatch[T_, M_]":
        """Creates a zeroed control input batch for the given horizon and rollout count."""
        array = np.zeros((horizon, BICYCLE_D_U, rollout_count))

        assert shape_of(array, matches=(horizon, BICYCLE_D_U, rollout_count))

        return NumPyBicycleControlInputBatch(array)

    @staticmethod
    def create[T_: int, M_: int](
        *, array: Array[Dims[T_, BicycleD_u, M_]]
    ) -> "NumPyBicycleControlInputBatch[T_, M_]":
        """Creates a NumPy bicycle control input batch from the given array."""

        return NumPyBicycleControlInputBatch(array)

    @staticmethod
    def of[T_: int](
        sequence: NumPyBicycleControlInputSequence[T_],
    ) -> "NumPyBicycleControlInputBatch[T_, D[1]]":
        """Creates a NumPy bicycle control input batch from a single control input sequence."""
        array = sequence.array[..., np.newaxis]

        assert shape_of(array, matches=(sequence.horizon, BICYCLE_D_U, 1))

        return NumPyBicycleControlInputBatch(array)

    def __array__(self, dtype: DataType | None = None) -> ControlInputBatchArray[T, M]:
        return self.array

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_u:
        return self.array.shape[1]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[2]

    @property
    def array(self) -> ControlInputBatchArray[T, M]:
        return self._array


@dataclass(frozen=True)
class NumPyBicycleObstacleStates[K: int]:
    array: Array[Dims[BicycleD_x, K]]

    @staticmethod
    def create(
        *,
        x: Array[Dims[K]],
        y: Array[Dims[K]],
        heading: Array[Dims[K]],
        speed: Array[Dims[K]],
    ) -> "NumPyBicycleObstacleStates[K]":
        """Creates a NumPy bicycle obstacle states from individual state components."""
        array = np.stack([x, y, heading, speed], axis=0)

        assert shape_of(array, matches=(BICYCLE_D_X, x.shape[0]))

        return NumPyBicycleObstacleStates(array)


@dataclass(frozen=True)
class NumPyBicycleObstacleStateSequences[T: int, K: int]:
    array: Array[Dims[T, BicycleD_x, K]]

    @staticmethod
    def create(
        *,
        x: Array[Dims[T, K]],
        y: Array[Dims[T, K]],
        heading: Array[Dims[T, K]],
        speed: Array[Dims[T, K]],
    ) -> "NumPyBicycleObstacleStateSequences[T, K]":
        """Creates a NumPy bicycle obstacle state sequences from individual state components."""
        T, K = x.shape
        array = np.stack([x, y, heading, speed], axis=1)

        assert shape_of(array, matches=(T, BICYCLE_D_X, K))

        return NumPyBicycleObstacleStateSequences(array)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, BicycleD_x, K]]:
        return self.array

    def single(self) -> Never:
        # TODO: Fix this!
        raise NotImplementedError(
            "single() is not implemented for NumPyBicycleObstacleStateSequences."
        )

    def x(self) -> Array[Dims[T, K]]:
        return self.array[:, 0, :]

    def y(self) -> Array[Dims[T, K]]:
        return self.array[:, 1, :]

    def heading(self) -> Array[Dims[T, K]]:
        return self.array[:, 2, :]

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return self.array.shape[1]

    @property
    def count(self) -> K:
        return self.array.shape[2]


@dataclass(frozen=True)
class NumPyBicycleObstacleVelocities[K: int]:
    steering_angles: Array[Dims[K]]

    @property
    def count(self) -> K:
        return self.steering_angles.shape[0]


@dataclass(frozen=True)
class NumPyBicycleObstacleControlInputSequences[T: int, K: int]:
    array: Array[Dims[T, BicycleD_u, K]]

    @staticmethod
    def create(
        *,
        accelerations: Array[Dims[T, K]],
        steering_angles: Array[Dims[T, K]],
    ) -> "NumPyBicycleObstacleControlInputSequences[T, K]":
        """Creates a NumPy bicycle obstacle control input sequences from individual input components."""
        T, K = accelerations.shape
        array = np.stack([accelerations, steering_angles], axis=1)

        assert shape_of(array, matches=(T, BICYCLE_D_U, K))

        return NumPyBicycleObstacleControlInputSequences(array)


@dataclass(kw_only=True, frozen=True)
class NumPyBicycleModel(
    DynamicalModel[
        NumPyBicycleState,
        NumPyBicycleStateSequence,
        NumPyBicycleStateBatch,
        NumPyBicycleControlInputSequence,
        NumPyBicycleControlInputBatch,
    ],
):
    _time_step_size: float
    wheelbase: float
    speed_limits: tuple[float, float]
    steering_limits: tuple[float, float]
    acceleration_limits: tuple[float, float]

    @staticmethod
    def create(
        *,
        time_step_size: float,
        wheelbase: float = 1.0,
        speed_limits: tuple[float, float] | None = None,
        steering_limits: tuple[float, float] | None = None,
        acceleration_limits: tuple[float, float] | None = None,
    ) -> "NumPyBicycleModel":
        """Creates a kinematic bicycle model that uses NumPy for computations."""

        return NumPyBicycleModel(
            _time_step_size=time_step_size,
            wheelbase=wheelbase,
            speed_limits=speed_limits if speed_limits is not None else NO_LIMITS,
            steering_limits=steering_limits
            if steering_limits is not None
            else NO_LIMITS,
            acceleration_limits=acceleration_limits
            if acceleration_limits is not None
            else NO_LIMITS,
        )

    def simulate[T: int, M: int](
        self,
        inputs: NumPyBicycleControlInputBatch[T, M],
        initial_state: NumPyBicycleState,
    ) -> NumPyBicycleStateBatch[T, M]:
        rollout_count = inputs.rollout_count

        initial = np.stack(
            [
                np.full(rollout_count, initial_state.x),
                np.full(rollout_count, initial_state.y),
                np.full(rollout_count, initial_state.heading),
                np.full(rollout_count, initial_state.speed),
            ]
        )

        return NumPyBicycleStateBatch(
            simulate(
                inputs.array,
                initial,
                time_step_size=self.time_step_size,
                wheelbase=self.wheelbase,
                speed_limits=self.speed_limits,
                steering_limits=self.steering_limits,
                acceleration_limits=self.acceleration_limits,
            )
        )

    def step[T: int](
        self, inputs: NumPyBicycleControlInputSequence[T], state: NumPyBicycleState
    ) -> NumPyBicycleState:
        state_as_rollouts = state.array.reshape(-1, 1)
        first_input = inputs.array[0].reshape(-1, 1)

        assert shape_of(
            state_as_rollouts, matches=(BICYCLE_D_X, 1), name="state reshaped for step"
        )
        assert shape_of(
            first_input,
            matches=(BICYCLE_D_U, 1),
            name="first control input reshaped for step",
        )

        return NumPyBicycleState(
            step(
                state_as_rollouts,
                first_input,
                time_step_size=self.time_step_size,
                wheelbase=self.wheelbase,
                speed_limits=self.speed_limits,
                steering_limits=self.steering_limits,
                acceleration_limits=self.acceleration_limits,
            )[:, 0]
        )

    def forward[T: int](
        self, inputs: NumPyBicycleControlInputSequence[T], state: NumPyBicycleState
    ) -> NumPyBicycleStateSequence[T]:
        return self.simulate(NumPyBicycleControlInputBatch.of(inputs), state).rollout(0)

    @property
    def time_step_size(self) -> float:
        return self._time_step_size


@dataclass(kw_only=True, frozen=True)
class NumPyBicycleObstacleModel(
    ObstacleModel[
        NumPyBicycleObstacleStatesHistory,
        NumPyBicycleObstacleStates,
        NumPyBicycleObstacleVelocities,
        NumPyBicycleObstacleControlInputSequences,
        NumPyBicycleObstacleStateSequences,
    ]
):
    time_step_size: float
    wheelbase: float

    @staticmethod
    def create(
        *, time_step_size: float, wheelbase: float = 1.0
    ) -> "NumPyBicycleObstacleModel":
        """Creates a NumPy bicycle obstacle model."""
        return NumPyBicycleObstacleModel(
            time_step_size=time_step_size, wheelbase=wheelbase
        )

    def estimate_state_from[K: int](
        self, history: NumPyBicycleObstacleStatesHistory[int, K]
    ) -> EstimatedObstacleStates[
        NumPyBicycleObstacleStates[K], NumPyBicycleObstacleVelocities[K]
    ]:
        assert history.horizon > 0, "History must have at least one time step."

        if history.horizon < 2:
            speeds = np.zeros((history.count))
            steering_angles = np.zeros((history.count))
        else:
            speeds = estimate_speeds_from(history, time_step_size=self.time_step_size)
            steering_angles = estimate_steering_angles_from(
                history,
                speeds=speeds,
                time_step_size=self.time_step_size,
                wheelbase=self.wheelbase,
            )

        assert shape_of(speeds, matches=(history.count,))
        assert shape_of(steering_angles, matches=(history.count,))

        return EstimatedObstacleStates(
            states=NumPyBicycleObstacleStates.create(
                x=history.x()[-1],
                y=history.y()[-1],
                heading=history.heading()[-1, :],
                speed=speeds,
            ),
            velocities=NumPyBicycleObstacleVelocities(steering_angles=steering_angles),
        )

    def input_to_maintain[K: int](
        self,
        velocities: NumPyBicycleObstacleVelocities[K],
        *,
        states: NumPyBicycleObstacleStates[K],
        horizon: int,
    ) -> NumPyBicycleObstacleControlInputSequences[int, K]:
        accelerations = np.zeros((horizon, velocities.count))

        assert shape_of(accelerations, matches=(horizon, velocities.count))

        return NumPyBicycleObstacleControlInputSequences.create(
            accelerations=accelerations,
            steering_angles=np.tile(
                velocities.steering_angles[np.newaxis, :], (horizon, 1)
            ),
        )

    def forward[T: int, K: int](
        self,
        *,
        current: NumPyBicycleObstacleStates[K],
        inputs: NumPyBicycleObstacleControlInputSequences[T, K],
    ) -> NumPyBicycleObstacleStateSequences[T, K]:
        result = simulate(
            inputs.array,
            current.array,
            time_step_size=self.time_step_size,
            wheelbase=self.wheelbase,
            speed_limits=(float("-inf"), float("inf")),
            steering_limits=(float("-inf"), float("inf")),
            acceleration_limits=(float("-inf"), float("inf")),
        )

        return NumPyBicycleObstacleStateSequences.create(
            x=result[:, 0, :],
            y=result[:, 1, :],
            heading=result[:, 2, :],
            speed=result[:, 3, :],
        )


def simulate[T: int, N: int](
    inputs: ControlInputBatchArray[T, N],
    initial: StatesAtTimeStep[N],
    *,
    time_step_size: float,
    wheelbase: float,
    speed_limits: tuple[float, float],
    steering_limits: tuple[float, float],
    acceleration_limits: tuple[float, float],
) -> StateBatchArray[T, N]:
    horizon = inputs.shape[0]
    rollout_count = inputs.shape[2]
    states = np.zeros((horizon, BICYCLE_D_X, rollout_count))
    current = initial

    for t in range(horizon):
        current = step(
            current,
            inputs[t],
            time_step_size=time_step_size,
            wheelbase=wheelbase,
            speed_limits=speed_limits,
            steering_limits=steering_limits,
            acceleration_limits=acceleration_limits,
        )
        states[t] = current

    assert shape_of(
        states, matches=(horizon, BICYCLE_D_X, rollout_count), name="simulated states"
    )

    return states


def step[M: int](
    state: StatesAtTimeStep[M],
    control: ControlInputsAtTimeStep[M],
    *,
    time_step_size: float,
    wheelbase: float,
    speed_limits: tuple[float, float],
    steering_limits: tuple[float, float],
    acceleration_limits: tuple[float, float],
) -> StatesAtTimeStep[M]:
    x, y, theta, v = state[0], state[1], state[2], state[3]
    a, delta = control[0], control[1]
    acceleration = np.clip(a, *acceleration_limits)
    steering = np.clip(delta, *steering_limits)

    new_x = x + v * np.cos(theta) * time_step_size
    new_y = y + v * np.sin(theta) * time_step_size
    new_theta = theta + v * np.tan(steering) / wheelbase * time_step_size
    new_v = np.clip(v + acceleration * time_step_size, *speed_limits)

    return np.stack([new_x, new_y, new_theta, new_v])


def estimate_speeds_from[K: int](
    history: NumPyBicycleObstacleStatesHistory[int, K], *, time_step_size: float
) -> Array[Dims[K]]:
    assert history.horizon >= 2, (
        "At least two history steps are required to estimate speed."
    )

    delta_x = history.x()[-1] - history.x()[-2]
    delta_y = history.y()[-1] - history.y()[-2]

    speeds = np.sqrt(delta_x**2 + delta_y**2) / time_step_size

    assert shape_of(speeds, matches=(history.count,), name="estimated speeds")

    return speeds


def estimate_steering_angles_from[K: int](
    history: NumPyBicycleObstacleStatesHistory[int, K],
    *,
    speeds: Array[Dims[K]],
    time_step_size: float,
    wheelbase: float,
) -> Array[Dims[K]]:
    assert history.horizon >= 2, (
        "At least two history steps are required to estimate heading."
    )

    heading_velocity = (
        history.heading()[-1, :] - history.heading()[-2, :]
    ) / time_step_size

    with np.errstate(invalid="ignore"):
        steering_angles = np.where(
            speeds > 1e-6, np.arctan(heading_velocity * wheelbase / speeds), 0.0
        )

    assert shape_of(
        steering_angles, matches=(history.count,), name="estimated steering angles"
    )

    return steering_angles



from dataclasses import dataclass
from functools import cached_property

from trajax.types.trajectories.common import (
    D_r,
    D_R,
    PathParameters,
    ReferencePoints,
    Positions,
    LateralPositions,
    LongitudinalPositions,
    Normals,
)

from numtypes import array, Array, Dims, D

import numpy as np


@dataclass(frozen=True)
class NumPyPathParameters[T: int, M: int](PathParameters[T, M]):
    array: Array[Dims[T, M]]

    def __array__(self) -> Array[Dims[T, M]]:
        return self.array

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[1]


@dataclass(frozen=True)
class NumPyPositions[T: int, M: int](Positions[T, M]):
    _x: Array[Dims[T, M]]
    _y: Array[Dims[T, M]]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Array[Dims[T_, M_]],
        y: Array[Dims[T_, M_]],
    ) -> "NumPyPositions[T_, M_]":
        """Creates a NumPy positions instance from x and y coordinate arrays."""
        return NumPyPositions(_x=x, _y=y)

    def __array__(self) -> Array[Dims[T, D[2], M]]:
        return self.array

    def x(self) -> Array[Dims[T, M]]:
        return self._x

    def y(self) -> Array[Dims[T, M]]:
        return self._y

    @property
    def horizon(self) -> T:
        return self._x.shape[0]

    @property
    def rollout_count(self) -> M:
        return self._x.shape[1]

    @property
    def array(self) -> Array[Dims[T, D[2], M]]:
        return self._array

    @cached_property
    def _array(self) -> Array[Dims[T, D[2], M]]:
        return np.stack([self._x, self._y], axis=1)


@dataclass(frozen=True)
class NumPyHeadings[T: int, M: int]:
    heading: Array[Dims[T, M]]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        heading: Array[Dims[T_, M_]],
    ) -> "NumPyHeadings[T_, M_]":
        """Creates a NumPy headings instance from an array of headings."""
        return NumPyHeadings(heading)


@dataclass(frozen=True)
class NumPyReferencePoints[T: int, M: int](ReferencePoints[T, M]):
    array: Array[Dims[T, D_r, M]]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Array[Dims[T_, M_]],
        y: Array[Dims[T_, M_]],
        heading: Array[Dims[T_, M_]],
    ) -> "NumPyReferencePoints[T_, M_]":
        """Creates a NumPy reference points instance from x, y, and heading arrays."""
        T, M = x.shape
        return NumPyReferencePoints(
            array=array(np.stack([x, y, heading], axis=1).tolist(), shape=(T, D_R, M))
        )

    def __array__(self) -> Array[Dims[T, D_r, M]]:
        return self.array

    def x(self) -> Array[Dims[T, M]]:
        return self.array[:, 0]

    def y(self) -> Array[Dims[T, M]]:
        return self.array[:, 1]

    def heading(self) -> Array[Dims[T, M]]:
        return self.array[:, 2]

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[2]

    @property
    def positions(self) -> Array[Dims[T, D[2], M]]:
        return self.array[:, :2]


@dataclass(frozen=True)
class NumPyLateralPositions[T: int, M: int](LateralPositions[T, M]):
    _array: Array[Dims[T, M]]

    @staticmethod
    def create[T_: int, M_: int](
        array: Array[Dims[T_, M_]],
    ) -> "NumPyLateralPositions[T_, M_]":
        """Creates a NumPy lateral positions instance from an array."""
        return NumPyLateralPositions(array)

    def __array__(self) -> Array[Dims[T, M]]:
        return self.array

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[1]

    @property
    def array(self) -> Array[Dims[T, M]]:
        return self._array


@dataclass(frozen=True)
class NumPyLongitudinalPositions[T: int, M: int](LongitudinalPositions[T, M]):
    _array: Array[Dims[T, M]]

    @staticmethod
    def create[T_: int, M_: int](
        array: Array[Dims[T_, M_]],
    ) -> "NumPyLongitudinalPositions[T_, M_]":
        """Creates a NumPy longitudinal positions instance from an array."""
        return NumPyLongitudinalPositions(array)

    def __array__(self) -> Array[Dims[T, M]]:
        return self.array

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[1]

    @property
    def array(self) -> Array[Dims[T, M]]:
        return self._array


@dataclass(frozen=True)
class NumPyNormals[T: int, M: int](Normals[T, M]):
    _array: Array[Dims[T, D[2], M]]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Array[Dims[T_, M_]],
        y: Array[Dims[T_, M_]],
    ) -> "NumPyNormals[T_, M_]":
        """Creates a NumPy normals instance from x and y coordinate arrays."""
        return NumPyNormals(np.stack([x, y], axis=1))

    def __array__(self) -> Array[Dims[T, D[2], M]]:
        return self.array

    def x(self) -> Array[Dims[T, M]]:
        return self.array[:, 0]

    def y(self) -> Array[Dims[T, M]]:
        return self.array[:, 1]

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self.array.shape[2]

    @property
    def array(self) -> Array[Dims[T, D[2], M]]:
        return self._array



from dataclasses import dataclass

from numtypes import Array, Dims, D, array

import numpy as np

type VerticesArray[P: int = int] = Array[Dims[P, D[2]]]


@dataclass(frozen=True)
class ConvexPolygon:
    """Describes a convex polygon for SAT-based distance computation.

    Vertices must be in counter-clockwise order in the local frame.
    """

    vertices: VerticesArray
    """The (x, y) coordinates of the polygon vertices in local frame, CCW order."""

    def __post_init__(self) -> None:
        verify_positive_edge_lengths(self.vertices)

    @staticmethod
    def square(*, size: float = 1.0) -> "ConvexPolygon":
        """A square centered at the origin, CCW order.

        Args:
            size: The length of each side of the square.
        """
        return ConvexPolygon(
            array(
                [[-0.5, -0.5], [0.5, -0.5], [0.5, 0.5], [-0.5, 0.5]],
                shape=(4, 2),
            )
            * size
        )

    @staticmethod
    def rectangle(*, length: float = 1.0, width: float = 1.0) -> "ConvexPolygon":
        """Rectangle centered at origin, CCW order. Length along x, width along y.

        Args:
            length: The length of the rectangle along the x-axis.
            width: The width of the rectangle along the y-axis.
        """
        half_length = length / 2
        half_width = width / 2
        return ConvexPolygon(
            array(
                [
                    [-half_length, -half_width],
                    [half_length, -half_width],
                    [half_length, half_width],
                    [-half_length, half_width],
                ],
                shape=(4, 2),
            )
        )


def verify_positive_edge_lengths(vertices: VerticesArray) -> None:
    edges = np.roll(vertices, -1, axis=0) - vertices
    edge_lengths = np.linalg.norm(edges, axis=-1)

    assert np.all(edge_lengths > 1e-10), (
        f"Polygon edges must have non-zero length. Got lengths: {edge_lengths}"
    )



from typing import Final, Sequence

from trajax.types import (
    SimulationDataAccessor,
    StateSequence,
    Control,
    Risk,
    ObstacleStates,
    StateTrajectories,
)


class access:
    states: Final = SimulationDataAccessor.create(StateSequence, key="states")
    controls: Final = SimulationDataAccessor.create(Sequence[Control], key="controls")
    risks: Final = SimulationDataAccessor.create(Sequence[Risk], key="risks")
    trajectories: Final = SimulationDataAccessor.create(
        Sequence[StateTrajectories], key="trajectories"
    )
    obstacle_states: Final = SimulationDataAccessor.create(
        ObstacleStates, key="obstacle_states"
    )
    obstacle_forecasts: Final = SimulationDataAccessor.create(
        Sequence[ObstacleStates], key="obstacle_forecasts"
    )



from typing import Self
from dataclasses import dataclass

from trajax.types import NumPyObstacleSimulator
from trajax.obstacles.basic import NumPyObstacleStatesForTimeStep

from numtypes import Array, Dims, D, shape_of

import numpy as np


@dataclass(kw_only=True)
class NumPyDynamicObstacleSimulator[K: int](
    NumPyObstacleSimulator[NumPyObstacleStatesForTimeStep[K]]
):
    last: NumPyObstacleStatesForTimeStep[K]
    velocities: Array[Dims[K, D[2]]]

    time_step: float | None = None

    @staticmethod
    def create[K_: int](
        *,
        positions: Array[Dims[K_, D[2]]],
        velocities: Array[Dims[K_, D[2]]],
    ) -> "NumPyDynamicObstacleSimulator[K_]":
        headings = headings_from(velocities)

        return NumPyDynamicObstacleSimulator(
            last=NumPyObstacleStatesForTimeStep.create(
                x=positions[:, 0], y=positions[:, 1], heading=headings
            ),
            velocities=velocities,
        )

    def with_time_step_size(self, time_step_size: float) -> Self:
        return self.__class__(
            last=self.last, velocities=self.velocities, time_step=time_step_size
        )

    def step(self) -> NumPyObstacleStatesForTimeStep[K]:
        assert self.time_step is not None, (
            "Time step must be set to advance obstacle states."
        )

        x, y = step_obstacles(
            x=self.last.x(),
            y=self.last.y(),
            velocities=self.velocities,
            time_step=self.time_step,
        )

        self.last = NumPyObstacleStatesForTimeStep.create(
            x=x, y=y, heading=self.last.heading()
        )

        return self.last

    @property
    def obstacle_count(self) -> int:
        return self.last.count


def step_obstacles[K: int](
    *,
    x: Array[Dims[K]],
    y: Array[Dims[K]],
    velocities: Array[Dims[K, D[2]]],
    time_step: float,
) -> tuple[Array[Dims[K]], Array[Dims[K]]]:
    K = x.shape[0]
    new_x = x + velocities[:, 0] * time_step
    new_y = y + velocities[:, 1] * time_step

    assert shape_of(new_x, matches=(K,))
    assert shape_of(new_y, matches=(K,))

    return new_x, new_y


def headings_from[K: int](velocities: Array[Dims[K, D[2]]]) -> Array[Dims[K]]:
    K = velocities.shape[0]
    speed = np.linalg.norm(velocities, axis=1)
    heading = np.where(
        speed > 1e-6, np.arctan2(velocities[:, 1], velocities[:, 0]), 0.0
    )

    assert shape_of(heading, matches=(K,))

    return heading



from .common import PredictingObstacleStateProvider as PredictingObstacleStateProvider
from .basic import (
    NumPySampledObstacleStates as NumPySampledObstacleStates,
    NumPyObstacleStates as NumPyObstacleStates,
    NumPyObstacleStatesForTimeStep as NumPyObstacleStatesForTimeStep,
    NumPyObstacle2dPositions as NumPyObstacle2dPositions,
    NumPyObstacle2dPositionsForTimeStep as NumPyObstacle2dPositionsForTimeStep,
)
from .accelerated import (
    JaxSampledObstacleStates as JaxSampledObstacleStates,
    JaxObstacleStates as JaxObstacleStates,
    JaxObstacleStatesForTimeStep as JaxObstacleStatesForTimeStep,
    JaxObstacle2dPositions as JaxObstacle2dPositions,
    JaxObstacle2dPositionsForTimeStep as JaxObstacle2dPositionsForTimeStep,
)
from .history import (
    NumPyObstacleIds as NumPyObstacleIds,
    NumPyObstacleStatesRunningHistory as NumPyObstacleStatesRunningHistory,
    JaxObstacleIds as JaxObstacleIds,
    JaxObstacleStatesRunningHistory as JaxObstacleStatesRunningHistory,
)
from .factory import obstacles as obstacles



from .basic import (
    NumPyBoundaryCost as NumPyBoundaryCost,
    NumPyFixedWidthBoundary as NumPyFixedWidthBoundary,
    NumPyPiecewiseFixedWidthBoundary as NumPyPiecewiseFixedWidthBoundary,
)
from .accelerated import (
    JaxBoundaryCost as JaxBoundaryCost,
    JaxFixedWidthBoundary as JaxFixedWidthBoundary,
    JaxPiecewiseFixedWidthBoundary as JaxPiecewiseFixedWidthBoundary,
)



from .common import (
    BoundaryPoints as BoundaryPoints,
    BoundaryDistance as BoundaryDistance,
    BoundaryDistanceExtractor as BoundaryDistanceExtractor,
    ExplicitBoundary as ExplicitBoundary,
    BoundaryWidthsDescription as BoundaryWidthsDescription,
    BoundaryWidths as BoundaryWidths,
)
from .basic import (
    NumPyBoundaryDistance as NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor as NumPyBoundaryDistanceExtractor,
)
from .accelerated import (
    JaxBoundaryDistance as JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor as JaxBoundaryDistanceExtractor,
)



from .common import (
    D_r as D_r,
    D_R as D_R,
    Trajectory as Trajectory,
    PathParameters as PathParameters,
    ReferencePoints as ReferencePoints,
    Positions as Positions,
    LateralPositions as LateralPositions,
    LongitudinalPositions as LongitudinalPositions,
    Normals as Normals,
)
from .basic import (
    NumPyPositions as NumPyPositions,
    NumPyHeadings as NumPyHeadings,
    NumPyPathParameters as NumPyPathParameters,
    NumPyReferencePoints as NumPyReferencePoints,
    NumPyLateralPositions as NumPyLateralPositions,
    NumPyLongitudinalPositions as NumPyLongitudinalPositions,
    NumPyNormals as NumPyNormals,
)
from .accelerated import (
    JaxPositions as JaxPositions,
    JaxHeadings as JaxHeadings,
    JaxPathParameters as JaxPathParameters,
    JaxReferencePoints as JaxReferencePoints,
    JaxLateralPositions as JaxLateralPositions,
    JaxLongitudinalPositions as JaxLongitudinalPositions,
    JaxNormals as JaxNormals,
)



from typing import Sequence, Callable, Any
from dataclasses import dataclass, field
from functools import cached_property

from trajax.types import (
    State,
    StateSequence,
    ControlInputSequence,
    Weights,
    DynamicalModel,
    Control,
    Mppi,
    StateTrajectories,
)
from trajax.collectors.access import access
from trajax.collectors.registry import (
    DataTransformer,
    IdentityTransformer,
    ModificationNotifierMixin,
    ListCollectorMixin,
    OnModifyCallback,
)


@dataclass(frozen=True)
class StateCollector[StateT = State, InputSequenceT = Any, WeightsT = Any](
    Mppi[StateT, InputSequenceT, WeightsT],
    ModificationNotifierMixin,
    ListCollectorMixin,
):
    inner: Mppi[StateT, InputSequenceT, WeightsT]
    transformer: DataTransformer[StateT, Any]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[StateT] = field(default_factory=list)

    @staticmethod
    def decorating[S: State, IS, W](
        mppi: Mppi[S, IS, W],
        *,
        transformer: DataTransformer[S, Any] = IdentityTransformer(),
    ) -> "StateCollector[S, IS, W]":
        return StateCollector(mppi, transformer=transformer)

    def step(
        self,
        *,
        temperature: float,
        nominal_input: InputSequenceT,
        initial_state: StateT,
    ) -> Control[InputSequenceT, WeightsT]:
        self._collected.append(initial_state)
        self.notify()

        return self.inner.step(
            temperature=temperature,
            nominal_input=nominal_input,
            initial_state=initial_state,
        )

    @property
    def key(self) -> str:
        return access.states.key


@dataclass(frozen=True)
class ControlCollector[
    StateT = Any,
    InputSequenceT = ControlInputSequence,
    WeightsT = Weights,
](
    Mppi[StateT, InputSequenceT, WeightsT],
    ModificationNotifierMixin,
    ListCollectorMixin,
):
    inner: Mppi[StateT, InputSequenceT, WeightsT]
    transformer: DataTransformer[Control[InputSequenceT, WeightsT]]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[Control[InputSequenceT, WeightsT]] = field(default_factory=list)

    @staticmethod
    def decorating[S, IS, W](
        mppi: Mppi[S, IS, W],
        *,
        transformer: DataTransformer[Control[IS, W]] = IdentityTransformer(),
    ) -> "ControlCollector[S, IS, W]":
        return ControlCollector(mppi, transformer=transformer)

    def step(
        self,
        *,
        temperature: float,
        nominal_input: InputSequenceT,
        initial_state: StateT,
    ) -> Control[InputSequenceT, WeightsT]:
        self._collected.append(
            control := self.inner.step(
                temperature=temperature,
                nominal_input=nominal_input,
                initial_state=initial_state,
            )
        )

        # TODO: Missing notify call.

        return control

    @property
    def key(self) -> str:
        return access.controls.key


@dataclass(frozen=True)
class TrajectoryCollector[
    StateT = Any,
    TrajectoriesT = StateSequence,
    InputSequenceT = Any,
    WeightsT = Any,
](Mppi[StateT, InputSequenceT, WeightsT], ModificationNotifierMixin):
    @dataclass(frozen=True)
    class LazyTrajectories[T]:
        factory: Callable[[], StateTrajectories[T]]

        @cached_property
        def value(self) -> StateTrajectories[T]:
            return self.factory()

    inner: Mppi[StateT, InputSequenceT, WeightsT]
    model: DynamicalModel[StateT, TrajectoriesT, Any, InputSequenceT, Any]
    transformer: DataTransformer[StateTrajectories[TrajectoriesT]]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[LazyTrajectories[TrajectoriesT]] = field(default_factory=list)

    @staticmethod
    def decorating[S, SS, IS, W](
        mppi: Mppi[S, IS, W],
        *,
        model: DynamicalModel[S, SS, Any, IS, Any],
        transformer: DataTransformer[StateTrajectories[SS]] = IdentityTransformer(),
    ) -> "TrajectoryCollector[S, SS, IS, W]":
        return TrajectoryCollector(mppi, model=model, transformer=transformer)

    def step(
        self,
        *,
        temperature: float,
        nominal_input: InputSequenceT,
        initial_state: StateT,
    ) -> Control[InputSequenceT, WeightsT]:
        control = self.inner.step(
            temperature=temperature,
            nominal_input=nominal_input,
            initial_state=initial_state,
        )

        self._collected.append(self._lazy_trajectories(control, initial_state))

        # TODO: Missing notify call.

        return control

    def collected(self, *, take: int) -> Sequence[StateTrajectories[TrajectoriesT]]:
        return [it.value for it in self._collected[:take]]

    @property
    def key(self) -> str:
        return access.trajectories.key

    @property
    def count(self) -> int:
        return len(self._collected)

    def _lazy_trajectories(
        self, control: Control[InputSequenceT, WeightsT], initial_state: StateT
    ) -> LazyTrajectories[TrajectoriesT]:
        def trajectories() -> StateTrajectories[TrajectoriesT]:
            return StateTrajectories(
                optimal=self.model.forward(control.optimal, state=initial_state),
                nominal=self.model.forward(control.nominal, state=initial_state),
            )

        return self.LazyTrajectories(trajectories)



try:
    from beartype import beartype  # pyright: ignore[reportMissingImports]

    typechecker = beartype
except ImportError:
    typechecker = None

from typing import Literal

from numtypes import Array
from jaxtyping import jaxtyped as jaxtyping_jaxtyped, Array as JaxArray

import jax
import numpy as np
import jax.numpy as jnp

type DataType = np.dtype
type Device = Literal["cpu", "default"]

jaxtyped = jaxtyping_jaxtyped(typechecker=typechecker)
"""Wrapper around `jaxtyping.jaxtyped` that conditionally applies type checking."""


def place(array: Array | JaxArray, *, device: Device) -> JaxArray:
    """Place an array on the specified device.

    Args:
        array: The array to place.
        device: The device to place the array on. Can be "cpu" or "default".

    Returns:
        The array placed on the specified device.
    """
    match device:
        case "cpu":
            return jnp.asarray(array, device=jax.devices("cpu")[0])
        case "default":
            return jnp.asarray(array)
        case _:
            raise ValueError(f"Unsupported device: {device}")



from typing import Final

from trajax.predictors.common import (
    StaticPredictor,
    CurvilinearPredictor,
    CovariancePadding,
)
from trajax.predictors.propagators import (
    NumPyLinearCovariancePropagator,
    JaxLinearCovariancePropagator,
)
from trajax.predictors.covariance import (
    NumPyConstantVarianceProvider,
    NumPyConstantCovarianceProvider,
    JaxConstantVarianceProvider,
    JaxConstantCovarianceProvider,
)


class predictor:
    class numpy:
        curvilinear: Final = CurvilinearPredictor.create
        static: Final = StaticPredictor.create

    class jax:
        curvilinear: Final = CurvilinearPredictor.create
        static: Final = StaticPredictor.create


class propagator:
    class numpy:
        linear: Final = NumPyLinearCovariancePropagator.create

        class covariance:
            constant_variance: Final = NumPyConstantVarianceProvider.create
            constant_covariance: Final = NumPyConstantCovarianceProvider.create

        padding: Final = CovariancePadding.create

    class jax:
        linear: Final = JaxLinearCovariancePropagator.create

        class covariance:
            constant_variance: Final = JaxConstantVarianceProvider.create
            constant_covariance: Final = JaxConstantCovarianceProvider.create

        padding: Final = CovariancePadding.create



from typing import Protocol, Final, Any

from numtypes import Array, Dims, D

D_R: Final = 3

type D_r = D[3]
"""Dimensionality of reference points (x, y, heading)."""


class Positions[T: int, M: int](Protocol):
    def __array__(self) -> Array[Dims[T, D[2], M]]:
        """Returns the positions as a NumPy array."""
        ...

    def x(self) -> Array[Dims[T, M]]:
        """Returns the x coordinates of the positions."""
        ...

    def y(self) -> Array[Dims[T, M]]:
        """Returns the y coordinates of the positions."""
        ...

    @property
    def horizon(self) -> T:
        """Returns the time horizon of the positions."""
        ...

    @property
    def rollout_count(self) -> M:
        """Returns the rollout count of the positions."""
        ...


class LateralPositions[T: int, M: int](Protocol):
    def __array__(self) -> Array[Dims[T, M]]:
        """Returns the lateral positions as a NumPy array."""
        ...

    @property
    def horizon(self) -> T:
        """Returns the time horizon of the lateral positions."""
        ...

    @property
    def rollout_count(self) -> M:
        """Returns the rollout count of the lateral positions."""
        ...


class LongitudinalPositions[T: int, M: int](Protocol):
    def __array__(self) -> Array[Dims[T, M]]:
        """Returns the longitudinal positions as a NumPy array."""
        ...

    @property
    def horizon(self) -> T:
        """Returns the time horizon of the longitudinal positions."""
        ...

    @property
    def rollout_count(self) -> M:
        """Returns the rollout count of the longitudinal positions."""
        ...


class PathParameters[T: int, M: int](Protocol):
    def __array__(self) -> Array[Dims[T, M]]:
        """Returns the path parameters as a NumPy array."""
        ...

    @property
    def horizon(self) -> T:
        """Returns the time horizon of the path parameters."""
        ...

    @property
    def rollout_count(self) -> M:
        """Returns the rollout count of the path parameters."""
        ...


class ReferencePoints[T: int, M: int](Protocol):
    def __array__(self) -> Array[Dims[T, D_r, M]]:
        """Returns the reference points as a NumPy array."""
        ...

    def x(self) -> Array[Dims[T, M]]:
        """Returns the x coordinates of the reference points."""
        ...

    def y(self) -> Array[Dims[T, M]]:
        """Returns the y coordinates of the reference points."""
        ...

    def heading(self) -> Array[Dims[T, M]]:
        """Returns the heading angles of the reference points."""
        ...

    @property
    def horizon(self) -> T:
        """Returns the time horizon of the reference points."""
        ...

    @property
    def rollout_count(self) -> M:
        """Returns the rollout count of the reference points."""
        ...


class Normals[T: int, M: int](Protocol):
    def __array__(self) -> Array[Dims[T, D[2], M]]:
        """Returns the normal vectors as a NumPy array."""
        ...

    def x(self) -> Array[Dims[T, M]]:
        """Returns the x components of the normal vectors."""
        ...

    def y(self) -> Array[Dims[T, M]]:
        """Returns the y components of the normal vectors."""
        ...

    @property
    def horizon(self) -> T:
        """Returns the time horizon of the normal vectors."""
        ...

    @property
    def rollout_count(self) -> M:
        """Returns the rollout count of the normal vectors."""
        ...


class Trajectory[
    PathParametersT,
    ReferencePointsT,
    PositionsT = Any,
    LateralT = Any,
    LongitudinalT = Any,
    NormalT = Any,
](Protocol):
    def query(self, parameters: PathParametersT) -> ReferencePointsT:
        """Queries the trajectory at the given path parameters, returning the reference points."""
        ...

    def lateral(self, positions: PositionsT) -> LateralT:
        """Computes the lateral deviation from the trajectory for each position."""
        ...

    def longitudinal(self, positions: PositionsT) -> LongitudinalT:
        """Computes the longitudinal coordinate along the trajectory for each position."""
        ...

    def normal(self, parameters: PathParametersT) -> NormalT:
        """Computes the normal vectors of the trajectory at the given path parameters."""
        ...

    @property
    def end(self) -> tuple[float, float]:
        """Returns the (x, y) position of the final point of the reference trajectory."""
        ...

    @property
    def path_length(self) -> float:
        """Returns the parameterization length of the reference trajectory.

        Note:
            This is not necessarily the same as the length of the trajectory itself, i.e.
            the integral of the arc length. For example, for a straight line from (0, 0) to (1, 1),
            the path length can be any number (e.g., 2.0, 5.0, 1e6, etc.), even though the actual
            length of the trajectory is sqrt(2).
        """
        ...

    @property
    def natural_length(self) -> float:
        """Returns the natural length of the trajectory (i.e., the integral of the arc length)."""
        ...



from typing import Sequence, cast
from dataclasses import dataclass
from functools import cached_property

from trajax.types import (
    jaxtyped,
    DataType,
    Device,
    place,
    D_o,
    D_O,
    SampledObstacleStates,
    ObstacleStates,
)
from trajax.obstacles.basic import NumPyObstacleStatesForTimeStep

from numtypes import Array, Dims, D
from jaxtyping import Array as JaxArray, Float

import numpy as np
import jax.numpy as jnp


type ObstacleCovarianceArray[T: int = int, K: int = int] = Float[
    JaxArray, f"T {D_O} {D_O} K"
]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxSampledObstacleStates[T: int, K: int, N: int](SampledObstacleStates[T, K, N]):
    _x: Float[JaxArray, "T K N"]
    _y: Float[JaxArray, "T K N"]
    _heading: Float[JaxArray, "T K N"]

    @staticmethod
    def create[T_: int, K_: int, N_: int](
        *,
        x: Float[JaxArray, "T K N"],
        y: Float[JaxArray, "T K N"],
        heading: Float[JaxArray, "T K N"],
        horizon: T_ | None = None,
        obstacle_count: K_ | None = None,
        sample_count: N_ | None = None,
    ) -> "JaxSampledObstacleStates[T_, K_, N_]":
        return JaxSampledObstacleStates(_x=x, _y=y, _heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_o, K, N]]:
        return self._numpy_array

    def x(self) -> Array[Dims[T, K, N]]:
        return np.asarray(self._x)

    def y(self) -> Array[Dims[T, K, N]]:
        return np.asarray(self._y)

    def heading(self) -> Array[Dims[T, K, N]]:
        return np.asarray(self._heading)

    def at(self, *, time_step: int, sample: int) -> "JaxObstacleStatesForTimeStep[K]":
        return JaxObstacleStatesForTimeStep.create(
            x=self._x[time_step, :, sample],
            y=self._y[time_step, :, sample],
            heading=self._heading[time_step, :, sample],
        )

    @property
    def horizon(self) -> T:
        return cast(T, self._x.shape[0])

    @property
    def dimension(self) -> D_o:
        return D_O

    @property
    def count(self) -> K:
        return cast(K, self._x.shape[1])

    @property
    def sample_count(self) -> N:
        return cast(N, self._x.shape[2])

    @property
    def x_array(self) -> Float[JaxArray, "T K N"]:
        return self._x

    @property
    def y_array(self) -> Float[JaxArray, "T K N"]:
        return self._y

    @property
    def heading_array(self) -> Float[JaxArray, "T K N"]:
        return self._heading

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, D_o, K, N]]:
        return np.stack([self._x, self._y, self._heading], axis=1)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacleStates[T: int, K: int](
    ObstacleStates[T, K, JaxSampledObstacleStates[T, K, D[1]]]
):
    _x: Float[JaxArray, "T K"]
    _y: Float[JaxArray, "T K"]
    _heading: Float[JaxArray, "T K"]
    _covariance: ObstacleCovarianceArray[T, K] | None = None

    @staticmethod
    def empty[T_: int, K_: int = D[0]](
        *, horizon: T_, obstacle_count: K_ = 0
    ) -> "JaxObstacleStates[T_, K_]":
        """Creates obstacle states for zero obstacles over the given time horizon."""
        empty = jnp.full((horizon, obstacle_count), fill_value=jnp.nan)

        return JaxObstacleStates.create(
            x=empty,
            y=empty,
            heading=empty,
            horizon=horizon,
            obstacle_count=obstacle_count,
        )

    @staticmethod
    def sampled[N: int](  # type: ignore
        *,
        x: Float[JaxArray, "T K N"],
        y: Float[JaxArray, "T K N"],
        heading: Float[JaxArray, "T K N"],
        sample_count: N | None = None,
    ) -> JaxSampledObstacleStates[T, K, N]:
        return JaxSampledObstacleStates.create(
            x=x, y=y, heading=heading, sample_count=sample_count
        )

    @staticmethod
    def wrap[T_: int, K_: int](
        array: Float[JaxArray, f"T {D_O} K"],
        *,
        horizon: T_ | None = None,
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacleStates[T_, K_]":
        return JaxObstacleStates.create(
            x=array[:, 0, :], y=array[:, 1, :], heading=array[:, 2, :]
        )

    @staticmethod
    def create[T_: int, K_: int](
        *,
        x: Float[JaxArray, "T K"],
        y: Float[JaxArray, "T K"],
        heading: Float[JaxArray, "T K"],
        covariance: ObstacleCovarianceArray[T_, K_] | None = None,
        horizon: T_ | None = None,
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacleStates[T_, K_]":
        return JaxObstacleStates(_x=x, _y=y, _heading=heading, _covariance=covariance)

    @staticmethod
    def of_states[T_: int, K_: int](
        obstacle_states: Sequence["JaxObstacleStatesForTimeStep[K_]"],
        *,
        horizon: T_ | None = None,
    ) -> "JaxObstacleStates[T_, K_]":
        assert len(obstacle_states) > 0, "Obstacle states sequence must not be empty."
        assert horizon is None or len(obstacle_states) == horizon, (
            f"Expected horizon {horizon}, but got {len(obstacle_states)} obstacle states."
        )

        K = max(states.count for states in obstacle_states)

        def pad(array: Float[JaxArray, "L"]) -> Float[JaxArray, "K"]:
            return jnp.pad(array, (0, K - len(array)), constant_values=jnp.nan)

        x = jnp.stack([pad(states.x_array) for states in obstacle_states])
        y = jnp.stack([pad(states.y_array) for states in obstacle_states])
        heading = jnp.stack([pad(states.heading_array) for states in obstacle_states])

        return JaxObstacleStates.create(x=x, y=y, heading=heading, horizon=horizon)

    @staticmethod
    def for_time_step[K_: int](
        *,
        x: Array[Dims[K_]] | Float[JaxArray, "K"],
        y: Array[Dims[K_]] | Float[JaxArray, "K"],
        heading: Array[Dims[K_]] | Float[JaxArray, "K"],
        obstacle_count: K_ | None = None,
        device: Device = "cpu",
    ) -> "JaxObstacleStatesForTimeStep[K_]":
        """Creates obstacle states for a single time step.

        Note:
            Since the common case is to further process this data on the CPU first,
            the default device is set to "cpu".
        """
        return JaxObstacleStatesForTimeStep.create(
            x=place(x, device=device),
            y=place(y, device=device),
            heading=place(heading, device=device),
            obstacle_count=obstacle_count,
        )

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_o, K]]:
        return np.stack([self._x, self._y, self._heading], axis=1)

    def x(self) -> Array[Dims[T, K]]:
        return np.asarray(self._x)

    def y(self) -> Array[Dims[T, K]]:
        return np.asarray(self._y)

    def heading(self) -> Array[Dims[T, K]]:
        return np.asarray(self._heading)

    def covariance(self) -> Array[Dims[T, D_o, D_o, K]] | None:
        return np.asarray(self._covariance) if self._covariance is not None else None

    def positions(self) -> "JaxObstacle2dPositions[T, K]":
        return JaxObstacle2dPositions.create(x=self._x, y=self._y)

    def single(self) -> JaxSampledObstacleStates[T, K, D[1]]:
        return JaxSampledObstacleStates.create(
            x=self._x[..., jnp.newaxis],
            y=self._y[..., jnp.newaxis],
            heading=self._heading[..., jnp.newaxis],
        )

    def last(self) -> "JaxObstacleStatesForTimeStep[K]":
        return self.at(time_step=self.horizon - 1)

    def at(self, time_step: int) -> "JaxObstacleStatesForTimeStep[K]":
        return JaxObstacleStatesForTimeStep.create(
            x=self._x[time_step],
            y=self._y[time_step],
            heading=self._heading[time_step],
        )

    @property
    def array(self) -> Float[JaxArray, "T D_o K"]:
        return jnp.stack([self._x, self._y, self._heading], axis=1)

    @property
    def x_array(self) -> Float[JaxArray, "T K"]:
        return self._x

    @property
    def y_array(self) -> Float[JaxArray, "T K"]:
        return self._y

    @property
    def heading_array(self) -> Float[JaxArray, "T K"]:
        return self._heading

    @property
    def covariance_array(self) -> ObstacleCovarianceArray[T, K] | None:
        return self._covariance

    @property
    def horizon(self) -> T:
        return cast(T, self._x.shape[0])

    @property
    def dimension(self) -> D_o:
        return D_O

    @property
    def count(self) -> K:
        return cast(K, self._x.shape[1])


@dataclass(kw_only=True, frozen=True)
class JaxObstacle2dPositions[T: int, K: int]:
    _x: Float[JaxArray, "T K"]
    _y: Float[JaxArray, "T K"]

    @staticmethod
    def create[T_: int, K_: int](
        *,
        x: Float[JaxArray, "T K"],
        y: Float[JaxArray, "T K"],
        horizon: T_ | None = None,
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacle2dPositions[T_, K_]":
        return JaxObstacle2dPositions(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D[2], K]]:
        return self._numpy_array

    @property
    def horizon(self) -> T:
        return cast(T, self._x.shape[0])

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> K:
        return cast(K, self._x.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T 2 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "T 2 K"]:
        return jnp.stack([self._x, self._y], axis=1)

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, D[2], K]]:
        return np.stack([np.asarray(self._x), np.asarray(self._y)], axis=1)


@dataclass(kw_only=True, frozen=True)
class JaxObstacle2dPositionsForTimeStep[K: int]:
    _x: Float[JaxArray, "K"]
    _y: Float[JaxArray, "K"]

    @staticmethod
    def create[K_: int](
        *,
        x: Float[JaxArray, "K"],
        y: Float[JaxArray, "K"],
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacle2dPositionsForTimeStep[K_]":
        return JaxObstacle2dPositionsForTimeStep(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D[2], K]]:
        return self._numpy_array

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> K:
        return cast(K, self._x.shape[0])

    @property
    def array(self) -> Float[JaxArray, "2 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "2 K"]:
        return jnp.stack([self._x, self._y], axis=0)

    @cached_property
    def _numpy_array(self) -> Array[Dims[D[2], K]]:
        return np.stack([np.asarray(self._x), np.asarray(self._y)], axis=0)


@dataclass(kw_only=True, frozen=True)
class JaxObstacleStatesForTimeStep[K: int]:
    _x: Float[JaxArray, "K"]
    _y: Float[JaxArray, "K"]
    _heading: Float[JaxArray, "K"]

    @staticmethod
    def create[K_: int](
        *,
        x: Float[JaxArray, "K"],
        y: Float[JaxArray, "K"],
        heading: Float[JaxArray, "K"],
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacleStatesForTimeStep[K_]":
        return JaxObstacleStatesForTimeStep(_x=x, _y=y, _heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D_o, K]]:
        return np.stack([self._x, self._y, self._heading], axis=0)

    def numpy(self) -> NumPyObstacleStatesForTimeStep[K]:
        return NumPyObstacleStatesForTimeStep.create(
            x=np.asarray(self._x),
            y=np.asarray(self._y),
            heading=np.asarray(self._heading),
        )

    def x(self) -> Array[Dims[K]]:
        return np.asarray(self._x)

    def y(self) -> Array[Dims[K]]:
        return np.asarray(self._y)

    def heading(self) -> Array[Dims[K]]:
        return np.asarray(self._heading)

    def positions(self) -> JaxObstacle2dPositionsForTimeStep[K]:
        return JaxObstacle2dPositionsForTimeStep.create(x=self._x, y=self._y)

    def replicate[T: int](self, *, horizon: T) -> JaxObstacleStates[T, K]:
        return JaxObstacleStates.create(
            x=jnp.tile(self._x[jnp.newaxis, :], (horizon, 1)),
            y=jnp.tile(self._y[jnp.newaxis, :], (horizon, 1)),
            heading=jnp.tile(self._heading[jnp.newaxis, :], (horizon, 1)),
        )

    @property
    def dimension(self) -> D_o:
        return D_O

    @property
    def count(self) -> K:
        return cast(K, self._x.shape[0])

    @property
    def x_array(self) -> Float[JaxArray, "K"]:
        return self._x

    @property
    def y_array(self) -> Float[JaxArray, "K"]:
        return self._y

    @property
    def heading_array(self) -> Float[JaxArray, "K"]:
        return self._heading

    @property
    def array(self) -> Float[JaxArray, f"{D_O} K"]:
        return jnp.stack([self._x, self._y, self._heading], axis=0)



from .basic import (
    NumPyInitialPositionCovariance as NumPyInitialPositionCovariance,
    NumPyInitialVelocityCovariance as NumPyInitialVelocityCovariance,
    NumPyPositionCovariance as NumPyPositionCovariance,
    NumPyInitialCovarianceProvider as NumPyInitialCovarianceProvider,
)
from .accelerated import (
    JaxInitialPositionCovariance as JaxInitialPositionCovariance,
    JaxInitialVelocityCovariance as JaxInitialVelocityCovariance,
    JaxPositionCovariance as JaxPositionCovariance,
    JaxInitialCovarianceProvider as JaxInitialCovarianceProvider,
)



from dataclasses import dataclass

from trajax.types import (
    CostFunction,
    ControlInputBatch,
    Trajectory,
    BoundaryPoints,
    BoundaryWidthsDescription,
    NumPyReferencePoints,
    NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor,
    NumPyCosts,
    NumPyPathParameters,
    NumPyPositions,
    NumPyLateralPositions,
    NumPyPositionExtractor,
)
from trajax.states import NumPySimpleCosts

from numtypes import Array, Dims

import numpy as np


@dataclass(frozen=True)
class NumPyBoundaryCost[StateT](CostFunction[ControlInputBatch, StateT, NumPyCosts]):
    distance: NumPyBoundaryDistanceExtractor[StateT, NumPyBoundaryDistance]
    distance_threshold: float
    weight: float

    @staticmethod
    def create[S](
        *,
        distance: NumPyBoundaryDistanceExtractor[S, NumPyBoundaryDistance],
        distance_threshold: float,
        weight: float,
    ) -> "NumPyBoundaryCost[S]":
        return NumPyBoundaryCost(
            distance=distance, distance_threshold=distance_threshold, weight=weight
        )

    def __call__[T: int, M: int](
        self, *, inputs: ControlInputBatch[T, int, M], states: StateT
    ) -> NumPyCosts[T, M]:
        cost = self.distance_threshold - self.distance(states=states).array

        return NumPySimpleCosts(self.weight * np.clip(cost, 0, None))


@dataclass(kw_only=True, frozen=True)
class NumPyFixedWidthBoundary[StateT](
    NumPyBoundaryDistanceExtractor[StateT, NumPyBoundaryDistance]
):
    reference: Trajectory[
        NumPyPathParameters, NumPyReferencePoints, NumPyPositions, NumPyLateralPositions
    ]
    position_extractor: NumPyPositionExtractor[StateT]
    _left: float
    _right: float

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[
            NumPyPathParameters,
            NumPyReferencePoints,
            NumPyPositions,
            NumPyLateralPositions,
        ],
        position_extractor: NumPyPositionExtractor[S],
        left: float,
        right: float,
    ) -> "NumPyFixedWidthBoundary[S]":
        """Creates a fixed-width boundary distance extractor.

        This component assumes a fixed-width corridor around a reference trajectory. The left and
        right widths can be different (asymmetric corridor).

        Args:
            reference: The reference trajectory defining the center of the corridor.
            position_extractor: Function to extract positions from states.
            left: The width of the left side of the corridor.
            right: The width of the right side of the corridor.
        """
        assert left >= -right, (
            f"The boundaries appear to be inverted. Left: {left}, Right: {right}. "
            f"Make sure the total width (left + right) is non-negative, got {left + right}."
        )

        return NumPyFixedWidthBoundary(
            reference=reference,
            position_extractor=position_extractor,
            _left=left,
            _right=right,
        )

    def __call__(self, *, states: StateT) -> NumPyBoundaryDistance:
        positions = self.position_extractor(states)
        lateral = self.reference.lateral(positions)

        distance_to_left = self._left + lateral.array
        distance_to_right = self._right - lateral.array

        return NumPyBoundaryDistance(np.minimum(distance_to_left, distance_to_right))

    def left[L: int](self, *, sample_count: L = 100) -> BoundaryPoints[L]:
        s = NumPyPathParameters(
            np.linspace(0, self.reference.path_length, sample_count).reshape(-1, 1)
        )

        return (
            self.reference.query(s).positions
            - self._left * self.reference.normal(s).array
        )[..., 0]

    def right[L: int](self, *, sample_count: L = 100) -> BoundaryPoints[L]:
        s = NumPyPathParameters(
            np.linspace(0, self.reference.path_length, sample_count).reshape(-1, 1)
        )

        return (
            self.reference.query(s).positions
            + self._right * self.reference.normal(s).array
        )[..., 0]


@dataclass(kw_only=True, frozen=True)
class NumPyPiecewiseFixedWidthBoundary[StateT, B: int = int](
    NumPyBoundaryDistanceExtractor[StateT, NumPyBoundaryDistance]
):
    reference: Trajectory[
        NumPyPathParameters, NumPyReferencePoints, NumPyPositions, NumPyLateralPositions
    ]
    position_extractor: NumPyPositionExtractor[StateT]
    breakpoints: Array[Dims[B]]
    left_widths: Array[Dims[B]]
    right_widths: Array[Dims[B]]

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[
            NumPyPathParameters,
            NumPyReferencePoints,
            NumPyPositions,
            NumPyLateralPositions,
        ],
        position_extractor: NumPyPositionExtractor[S],
        widths: BoundaryWidthsDescription,
    ) -> "NumPyPiecewiseFixedWidthBoundary[S]":
        """Creates a piecewise fixed-width boundary distance extractor.

        This component assumes a piecewise constant corridor width around a reference trajectory.
        Different segments of the trajectory can have different left and right widths.

        Args:
            reference: The reference trajectory defining the center of the corridor.
            position_extractor: Function to extract positions from states.
            widths: A mapping from longitudinal breakpoints to {"left": float, "right": float}
                   dictionaries defining the corridor widths for each segment.
        """
        sorted_breakpoints = sorted(widths.keys())

        return NumPyPiecewiseFixedWidthBoundary(
            reference=reference,
            position_extractor=position_extractor,
            breakpoints=np.array(sorted_breakpoints),
            left_widths=np.array([widths[s]["left"] for s in sorted_breakpoints]),
            right_widths=np.array([widths[s]["right"] for s in sorted_breakpoints]),
        )

    def __call__(self, *, states: StateT) -> NumPyBoundaryDistance:
        positions = self.position_extractor(states)
        lateral = self.reference.lateral(positions)
        longitudinal = self.reference.longitudinal(positions)

        segment_indices = (
            np.searchsorted(self.breakpoints, longitudinal.array, side="right") - 1
        )
        segment_indices = np.clip(segment_indices, 0, len(self.breakpoints) - 1)

        left = self.left_widths[segment_indices]
        right = self.right_widths[segment_indices]

        distance_to_left = left + lateral.array
        distance_to_right = right - lateral.array

        return NumPyBoundaryDistance(np.minimum(distance_to_left, distance_to_right))

    def left[L: int](self, *, sample_count: L = 100) -> BoundaryPoints[L]:
        s_values = np.linspace(0, self.reference.path_length, sample_count)
        s = NumPyPathParameters(s_values.reshape(-1, 1))

        segment_indices = np.searchsorted(self.breakpoints, s_values, side="right") - 1
        segment_indices = np.clip(segment_indices, 0, len(self.breakpoints) - 1)
        left = self.left_widths[segment_indices]

        return (
            self.reference.query(s).positions
            - left[:, np.newaxis, np.newaxis] * self.reference.normal(s).array
        )[..., 0]

    def right[L: int](self, *, sample_count: L = 100) -> BoundaryPoints[L]:
        s_values = np.linspace(0, self.reference.path_length, sample_count)
        s = NumPyPathParameters(s_values.reshape(-1, 1))

        segment_indices = np.searchsorted(self.breakpoints, s_values, side="right") - 1
        segment_indices = np.clip(segment_indices, 0, len(self.breakpoints) - 1)
        right = self.right_widths[segment_indices]

        return (
            self.reference.query(s).positions
            + right[:, np.newaxis, np.newaxis] * self.reference.normal(s).array
        )[..., 0]



from typing import Any, overload, cast
from dataclasses import dataclass

from trajax.types import (
    jaxtyped,
    DataType,
    StateBatch,
    ControlInputBatch,
    CostFunction,
    ContouringCost,
    LagCost,
    Error,
    Trajectory,
    JaxControlInputBatch,
    JaxCosts,
    JaxPathParameters,
    JaxReferencePoints,
    JaxPositions,
    JaxPositionExtractor,
    JaxPathParameterExtractor,
    JaxPathVelocityExtractor,
)
from trajax.states import JaxSimpleCosts

from jaxtyping import Array as JaxArray, Float, Scalar
from numtypes import Array, Dims

import numpy as np
import jax
import jax.numpy as jnp


@jaxtyped
@dataclass(frozen=True)
class JaxError[T: int, M: int](Error[T, M]):
    array: Float[JaxArray, "T M"]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        return np.asarray(self.array)


@dataclass(kw_only=True, frozen=True)
class JaxContouringCost[StateBatchT](
    ContouringCost[ControlInputBatch, StateBatchT, JaxError],
    CostFunction[ControlInputBatch, StateBatchT, JaxCosts],
):
    reference: Trajectory[JaxPathParameters, JaxReferencePoints]
    path_parameter_extractor: JaxPathParameterExtractor[StateBatchT]
    position_extractor: JaxPositionExtractor[StateBatchT]
    weight: float

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[JaxPathParameters, JaxReferencePoints],
        path_parameter_extractor: JaxPathParameterExtractor[S],
        position_extractor: JaxPositionExtractor[S],
        weight: float,
    ) -> "JaxContouringCost[S]":
        """Creates a contouring cost implemented with JAX.

        Args:
            reference: The reference trajectory to follow.
            path_parameter_extractor: Extracts the path parameters from a state batch.
            position_extractor: Extracts the (x, y) positions from a state batch.
            weight: The weight of the contouring cost.
        """
        return JaxContouringCost(
            reference=reference,
            path_parameter_extractor=path_parameter_extractor,
            position_extractor=position_extractor,
            weight=weight,
        )

    def __call__[T: int, M: int](
        self, *, inputs: ControlInputBatch[T, int, M], states: StateBatchT
    ) -> JaxCosts[T, M]:
        ref_points = self.reference.query(self.path_parameter_extractor(states))
        heading = ref_points.heading_array
        positions = self.position_extractor(states)

        return JaxSimpleCosts(
            contour_cost(
                heading=heading,
                x=positions.x_array,
                y=positions.y_array,
                x_ref=ref_points.x_array,
                y_ref=ref_points.y_array,
                weight=self.weight,
            )
        )

    def error[T: int = int, M: int = int](
        self, *, states: StateBatchT
    ) -> JaxError[T, M]:
        ref_points = self.reference.query(self.path_parameter_extractor(states))
        heading = ref_points.heading_array
        positions = self.position_extractor(states)

        return JaxError(
            contour_error(
                heading=heading,
                x=positions.x_array,
                y=positions.y_array,
                x_ref=ref_points.x_array,
                y_ref=ref_points.y_array,
            )
        )


@dataclass(kw_only=True, frozen=True)
class JaxLagCost[StateBatchT](
    LagCost[ControlInputBatch, StateBatchT, JaxError],
    CostFunction[ControlInputBatch, StateBatchT, JaxCosts],
):
    reference: Trajectory[JaxPathParameters, JaxReferencePoints]
    path_parameter_extractor: JaxPathParameterExtractor[StateBatchT]
    position_extractor: JaxPositionExtractor[StateBatchT]
    weight: float

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[JaxPathParameters, JaxReferencePoints],
        path_parameter_extractor: JaxPathParameterExtractor[S],
        position_extractor: JaxPositionExtractor[S],
        weight: float,
    ) -> "JaxLagCost[S]":
        """Creates a lag cost implemented with JAX.

        Args:
            reference: The reference trajectory to follow.
            path_parameter_extractor: Extracts the path parameters from a state batch.
            position_extractor: Extracts the (x, y) positions from a state batch.
            weight: The weight of the lag cost.
        """
        return JaxLagCost(
            reference=reference,
            path_parameter_extractor=path_parameter_extractor,
            position_extractor=position_extractor,
            weight=weight,
        )

    def __call__[T: int, M: int](
        self, *, inputs: ControlInputBatch[T, int, M], states: StateBatchT
    ) -> JaxCosts[T, M]:
        ref_points, positions = self._reference_points_and_positions(states=states)
        heading = ref_points.heading_array

        return JaxSimpleCosts(
            lag_cost(
                heading=heading,
                x=positions.x_array,
                y=positions.y_array,
                x_ref=ref_points.x_array,
                y_ref=ref_points.y_array,
                weight=self.weight,
            )
        )

    def error[T: int = int, M: int = int](
        self, *, states: StateBatchT
    ) -> JaxError[T, M]:
        ref_points, positions = self._reference_points_and_positions(states=states)
        heading = ref_points.heading_array

        return JaxError(
            lag_error(
                heading=heading,
                x=positions.x_array,
                y=positions.y_array,
                x_ref=ref_points.x_array,
                y_ref=ref_points.y_array,
            )
        )

    def _reference_points_and_positions[T: int, M: int](
        self, *, states: StateBatchT
    ) -> tuple[JaxReferencePoints[T, M], JaxPositions[T, M]]:
        return (
            self.reference.query(self.path_parameter_extractor(states)),
            self.position_extractor(states),
        )


@dataclass(kw_only=True, frozen=True)
class JaxProgressCost[InputBatchT](CostFunction[InputBatchT, StateBatch, JaxCosts]):
    path_velocity_extractor: JaxPathVelocityExtractor[InputBatchT]
    time_step_size: float
    weight: float

    @staticmethod
    def create[I](
        *,
        path_velocity_extractor: JaxPathVelocityExtractor[I],
        time_step_size: float,
        weight: float,
    ) -> "JaxProgressCost[I]":
        """Creates a progress cost implemented with JAX.

        Args:
            path_velocity_extractor: Extracts path velocities from a control input batch.
            time_step_size: The time step size used in the trajectory simulation.
            weight: The weight of the progress cost.
        """
        return JaxProgressCost(
            path_velocity_extractor=path_velocity_extractor,
            time_step_size=time_step_size,
            weight=weight,
        )

    def __call__[T: int, M: int](
        self, *, inputs: InputBatchT, states: StateBatch[T, int, M]
    ) -> JaxCosts[T, M]:
        path_velocities = self.path_velocity_extractor(inputs)

        return JaxSimpleCosts(
            progress_cost(
                path_velocities=path_velocities,
                time_step_size=self.time_step_size,
                weight=self.weight,
            )
        )


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxControlSmoothingCost[D_u: int](
    CostFunction[JaxControlInputBatch[int, D_u, int], StateBatch, JaxCosts]
):
    weights: Float[JaxArray, "D_u"]
    dimensions: D_u

    @overload
    @staticmethod
    def create[D_u_: int](
        *, weights: Array[Dims[D_u_]]
    ) -> "JaxControlSmoothingCost[D_u_]":
        """Creates a control smoothing cost implemented with JAX.

        Args:
            weights: The weights for each control input dimension.
        """
        ...

    @overload
    @staticmethod
    def create[D_u_: int](
        *, weights: Float[JaxArray, "D_u_"], dimensions: D_u_ | None = None
    ) -> "JaxControlSmoothingCost[D_u_]":
        """Creates a control smoothing cost implemented with JAX.

        Args:
            weights: The weights for each control input dimension.
            dimensions: The number of control input dimensions.
        """
        ...

    @staticmethod
    def create[D_u_: int](
        *,
        weights: Array[Dims[D_u_]] | Float[JaxArray, "D_u_"],
        dimensions: D_u_ | None = None,
    ) -> "JaxControlSmoothingCost[D_u_]":
        dimensions = (
            dimensions if dimensions is not None else cast(D_u_, weights.shape[0])
        )
        return JaxControlSmoothingCost(
            weights=jnp.asarray(weights), dimensions=dimensions
        )

    def __call__[T: int, M: int](
        self,
        *,
        inputs: JaxControlInputBatch[T, D_u, M],
        states: StateBatch[T, int, M],
    ) -> JaxCosts[T, M]:
        return JaxSimpleCosts(
            control_smoothing_cost(inputs=inputs.array, weights=self.weights)
        )


@dataclass(kw_only=True, frozen=True)
class JaxControlEffortCost[D_u: int](
    CostFunction[JaxControlInputBatch[int, D_u, int], Any, JaxCosts]
):
    weights: Float[JaxArray, "D_u"]

    @staticmethod
    def create[D_u_: int](
        *,
        weights: Float[JaxArray, "D_u_"] | Array[Dims[D_u_]],
    ) -> "JaxControlEffortCost[D_u_]":
        """Creates a control effort cost implemented with JAX.

        Args:
            weights: The weights for each control input dimension.
        """
        return JaxControlEffortCost(weights=jnp.asarray(weights))

    def __call__[T: int, M: int](
        self, *, inputs: JaxControlInputBatch[T, D_u, M], states: Any
    ) -> JaxCosts[T, M]:
        return JaxSimpleCosts(
            control_effort_cost(inputs=inputs.array, weights=self.weights)
        )


@jax.jit
@jaxtyped
def contour_error(
    *,
    heading: Float[JaxArray, "T M"],
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    x_ref: Float[JaxArray, "T M"],
    y_ref: Float[JaxArray, "T M"],
) -> Float[JaxArray, "T M"]:
    return jnp.sin(heading) * (x - x_ref) - jnp.cos(heading) * (y - y_ref)


@jax.jit
@jaxtyped
def contour_cost(
    *,
    heading: Float[JaxArray, "T M"],
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    x_ref: Float[JaxArray, "T M"],
    y_ref: Float[JaxArray, "T M"],
    weight: Scalar,
) -> Float[JaxArray, "T M"]:
    error = contour_error(heading=heading, x=x, y=y, x_ref=x_ref, y_ref=y_ref)
    return weight * error**2


@jax.jit
@jaxtyped
def lag_error(
    *,
    heading: Float[JaxArray, "T M"],
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    x_ref: Float[JaxArray, "T M"],
    y_ref: Float[JaxArray, "T M"],
) -> Float[JaxArray, "T M"]:
    return -jnp.cos(heading) * (x - x_ref) - jnp.sin(heading) * (y - y_ref)


@jax.jit
@jaxtyped
def lag_cost(
    *,
    heading: Float[JaxArray, "T M"],
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    x_ref: Float[JaxArray, "T M"],
    y_ref: Float[JaxArray, "T M"],
    weight: Scalar,
) -> Float[JaxArray, "T M"]:
    error = lag_error(heading=heading, x=x, y=y, x_ref=x_ref, y_ref=y_ref)
    return weight * error**2


@jax.jit
@jaxtyped
def progress_cost(
    *, path_velocities: Float[JaxArray, "T M"], time_step_size: Scalar, weight: Scalar
) -> Float[JaxArray, "T M"]:
    return -weight * path_velocities * time_step_size


@jax.jit
@jaxtyped
def control_smoothing_cost(
    *, inputs: Float[JaxArray, "T D_u M"], weights: Float[JaxArray, " D_u"]
) -> Float[JaxArray, "T M"]:
    diffs = jnp.diff(inputs, axis=0, prepend=inputs[0:1, :, :])
    squared_diffs = diffs**2
    weighted_squared_diffs = squared_diffs * weights[jnp.newaxis, :, jnp.newaxis]
    cost_per_time_step = jnp.sum(weighted_squared_diffs, axis=1)
    return cost_per_time_step


@jax.jit
@jaxtyped
def control_effort_cost(
    *, inputs: Float[JaxArray, "T D_u M"], weights: Float[JaxArray, "D_u"]
) -> Float[JaxArray, "T M"]:
    return jnp.einsum("u,tum->tm", weights, inputs**2)



from typing import Protocol, cast
from dataclasses import dataclass
from functools import cached_property

from trajax.types.array import DataType, jaxtyped
from trajax.types.costs.boundary.common import BoundaryDistanceExtractor

from numtypes import Array, Dims
from jaxtyping import Array as JaxArray, Float

import numpy as np


class JaxBoundaryDistanceExtractor[StateBatchT, DistanceT](
    BoundaryDistanceExtractor[StateBatchT, DistanceT], Protocol
): ...


@jaxtyped
@dataclass(frozen=True)
class JaxBoundaryDistance[T: int, M: int]:
    _array: Float[JaxArray, "T M"]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        return self._numpy_array

    @property
    def horizon(self) -> T:
        return cast(T, self._array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self._array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, M]]:
        return np.array(self._array)



from typing import Final

from trajax.types import NumPyRisk, NumPyRiskMetric, JaxRisk, JaxRiskMetric
from trajax.costs.collision import NumPyNoMetric, JaxNoMetric
from trajax.costs.risk.base import RisKitRiskMetric

import riskit as rk


class risk:
    class numpy:
        none: Final = NumPyNoMetric.create

        @staticmethod
        def expected_value(*, sample_count: int) -> NumPyRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.numpy.create(),
                creator=lambda *, cost, backend: rk.risk.expected_value_of(
                    cost, backend=backend
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=NumPyRisk,
                name="Expected Value",
            )

        @staticmethod
        def mean_variance(*, gamma: float, sample_count: int) -> NumPyRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.numpy.create(),
                creator=lambda *, cost, backend: rk.risk.mean_variance_of(
                    cost, backend=backend, gamma=gamma
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=NumPyRisk,
                name="Mean-Variance",
            )

        @staticmethod
        def var(*, alpha: float, sample_count: int) -> NumPyRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.numpy.create(),
                creator=lambda *, cost, backend: rk.risk.var_of(
                    cost, backend=backend, alpha=alpha
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=NumPyRisk,
                name="Value at Risk",
            )

        @staticmethod
        def cvar(*, alpha: float, sample_count: int) -> NumPyRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.numpy.create(),
                creator=lambda *, cost, backend: rk.risk.cvar_of(
                    cost, backend=backend, alpha=alpha
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=NumPyRisk,
                name="Conditional Value at Risk",
            )

        @staticmethod
        def entropic_risk(*, theta: float, sample_count: int) -> NumPyRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.numpy.create(),
                creator=lambda *, cost, backend: rk.risk.entropic_risk_of(
                    cost, backend=backend, theta=theta
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=NumPyRisk,
                name="Entropic Risk",
            )

    class jax:
        none: Final = JaxNoMetric.create

        @staticmethod
        def expected_value(*, sample_count: int) -> JaxRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.jax.create(),
                creator=lambda *, cost, backend: rk.risk.expected_value_of(
                    cost, backend=backend
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=JaxRisk,
                name="Expected Value",
            )

        @staticmethod
        def mean_variance(*, gamma: float, sample_count: int) -> JaxRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.jax.create(),
                creator=lambda *, cost, backend: rk.risk.mean_variance_of(
                    cost, backend=backend, gamma=gamma
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=JaxRisk,
                name="Mean-Variance",
            )

        @staticmethod
        def var(*, alpha: float, sample_count: int) -> JaxRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.jax.create(),
                creator=lambda *, cost, backend: rk.risk.var_of(
                    cost, backend=backend, alpha=alpha
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=JaxRisk,
                name="Value at Risk",
            )

        @staticmethod
        def cvar(*, alpha: float, sample_count: int) -> JaxRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.jax.create(),
                creator=lambda *, cost, backend: rk.risk.cvar_of(
                    cost, backend=backend, alpha=alpha
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=JaxRisk,
                name="Conditional Value at Risk",
            )

        @staticmethod
        def entropic_risk(*, theta: float, sample_count: int) -> JaxRiskMetric:
            return RisKitRiskMetric.create(
                backend=rk.backend.jax.create(),
                creator=lambda *, cost, backend: rk.risk.entropic_risk_of(
                    cost, backend=backend, theta=theta
                ).sampled_with(rk.sampler.monte_carlo(sample_count)),
                to_risk=JaxRisk,
                name="Entropic Risk",
            )



from typing import cast, overload
from dataclasses import dataclass
from functools import cached_property

from trajax.types.array import jaxtyped
from trajax.types.trajectories.common import (
    D_r,
    D_R,
    PathParameters,
    ReferencePoints,
    Positions,
    LateralPositions,
    LongitudinalPositions,
    Normals,
)

from jaxtyping import Array as JaxArray, Float
from numtypes import Array, Dims, D

import jax.numpy as jnp
import numpy as np


@dataclass(frozen=True)
class JaxPathParameters[T: int, M: int](PathParameters[T, M]):
    array: Float[JaxArray, "T M"]

    @overload
    @staticmethod
    def create[T_: int, M_: int](
        array: Array[Dims[T_, M_]],
    ) -> "JaxPathParameters[T_, M_]":
        """Creates a JAX path parameters instance from a NumPy array."""
        ...

    @overload
    @staticmethod
    def create[T_: int, M_: int](
        array: Float[JaxArray, "T M"],
        *,
        horizon: T_ | None = None,
        rollout_count: M_ | None = None,
    ) -> "JaxPathParameters[T_, M_]":
        """Creates a JAX path parameters instance from a JAX array."""
        ...

    @staticmethod
    def create[T_: int, M_: int](
        array: Array[Dims[T_, M_]] | Float[JaxArray, "T M"],
        *,
        horizon: T_ | None = None,
        rollout_count: M_ | None = None,
    ) -> "JaxPathParameters[T_, M_]":
        return JaxPathParameters(array=jnp.asarray(array))

    def __array__(self) -> Array[Dims[T, M]]:
        return np.asarray(self.array)

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self.array.shape[1])


@jaxtyped
@dataclass(frozen=True)
class JaxPositions[T: int, M: int](Positions[T, M]):
    x_array: Float[JaxArray, "T M"]
    y_array: Float[JaxArray, "T M"]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Float[JaxArray, "T M"] | Array[Dims[T_, M_]],
        y: Float[JaxArray, "T M"] | Array[Dims[T_, M_]],
        horizon: T_ | None = None,
        rollout_count: M_ | None = None,
    ) -> "JaxPositions[T_, M_]":
        """Creates a JAX positions instance from x and y coordinate arrays."""
        x_jax = jnp.asarray(x)
        y_jax = jnp.asarray(y)
        horizon = horizon if horizon is not None else cast(T_, x_jax.shape[0])
        rollout_count = (
            rollout_count if rollout_count is not None else cast(M_, x_jax.shape[1])
        )

        assert x_jax.shape == y_jax.shape == (horizon, rollout_count), (
            f"Expected x and y to have shape {(horizon, rollout_count)}, "
            f"but got {x_jax.shape} and {y_jax.shape}."
        )

        return JaxPositions(x_array=x_jax, y_array=y_jax)

    def __array__(self) -> Array[Dims[T, D[2], M]]:
        return self._numpy_array

    def x(self) -> Array[Dims[T, M]]:
        return np.asarray(self.x_array)

    def y(self) -> Array[Dims[T, M]]:
        return np.asarray(self.y_array)

    @property
    def horizon(self) -> T:
        return cast(T, self.x_array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self.x_array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "T 2 M"]:
        return jnp.stack([self.x_array, self.y_array], axis=1)

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, D[2], M]]:
        return np.asarray(jnp.stack([self.x_array, self.y_array], axis=1))


@jaxtyped
@dataclass(frozen=True)
class JaxHeadings[T: int, M: int]:
    heading: Float[JaxArray, "T M"]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        heading: Float[JaxArray, "T M"],
        horizon: T_ | None = None,
        rollout_count: M_ | None = None,
    ) -> "JaxHeadings[T_, M_]":
        """Creates a JAX headings instance from an array of headings."""
        return JaxHeadings(heading)


@dataclass(frozen=True)
class JaxReferencePoints[T: int, M: int](ReferencePoints[T, M]):
    array: Float[JaxArray, f"T {D_R} M"]

    @overload
    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Array[Dims[T_, M_]],
        y: Array[Dims[T_, M_]],
        heading: Array[Dims[T_, M_]],
    ) -> "JaxReferencePoints[T_, M_]":
        """Creates a JAX reference points instance from NumPy arrays."""
        ...

    @overload
    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Float[JaxArray, "T M"],
        y: Float[JaxArray, "T M"],
        heading: Float[JaxArray, "T M"],
        horizon: T_,
        rollout_count: M_,
    ) -> "JaxReferencePoints[T_, M_]":
        """Creates a JAX reference points instance from JAX arrays."""
        ...

    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Array[Dims[T_, M_]] | Float[JaxArray, "T M"],
        y: Array[Dims[T_, M_]] | Float[JaxArray, "T M"],
        heading: Array[Dims[T_, M_]] | Float[JaxArray, "T M"],
        horizon: T_ | None = None,
        rollout_count: M_ | None = None,
    ) -> "JaxReferencePoints[T_, M_]":
        horizon = horizon if horizon is not None else cast(T_, x.shape[0])
        rollout_count = (
            rollout_count if rollout_count is not None else cast(M_, x.shape[1])
        )

        assert x.shape == y.shape == heading.shape == (horizon, rollout_count), (
            f"Expected x, y, and heading to have shape {(horizon, rollout_count)}, "
            f"but got {x.shape}, {y.shape}, and {heading.shape}."
        )

        return JaxReferencePoints(array=jnp.stack([x, y, heading], axis=1))

    def __array__(self) -> Array[Dims[T, D_r, M]]:
        return np.asarray(self.array)

    def x(self) -> Array[Dims[T, M]]:
        return np.asarray(self.array[:, 0])

    def y(self) -> Array[Dims[T, M]]:
        return np.asarray(self.array[:, 1])

    def heading(self) -> Array[Dims[T, M]]:
        return np.asarray(self.array[:, 2])

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self.array.shape[2])

    @property
    def x_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 0]

    @property
    def y_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 1]

    @property
    def heading_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 2]

    @property
    def positions_array(self) -> Float[JaxArray, "T 2 M"]:
        return self.array[:, :2]


@jaxtyped
@dataclass(frozen=True)
class JaxLateralPositions[T: int, M: int](LateralPositions[T, M]):
    _array: Float[JaxArray, "T M"]

    @staticmethod
    def create[T_: int, M_: int](
        array: Float[JaxArray, "T M"] | Array[Dims[T_, M_]],
    ) -> "JaxLateralPositions[T_, M_]":
        """Creates a JAX lateral positions instance from an array."""
        return JaxLateralPositions(jnp.asarray(array))

    def __array__(self) -> Array[Dims[T, M]]:
        return self._numpy_array

    @property
    def horizon(self) -> T:
        return cast(T, self._array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self._array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, M]]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxLongitudinalPositions[T: int, M: int](LongitudinalPositions[T, M]):
    _array: Float[JaxArray, "T M"]

    @staticmethod
    def create[T_: int, M_: int](
        array: Float[JaxArray, "T M"] | Array[Dims[T_, M_]],
    ) -> "JaxLongitudinalPositions[T_, M_]":
        """Creates a JAX longitudinal positions instance from an array."""
        return JaxLongitudinalPositions(jnp.asarray(array))

    def __array__(self) -> Array[Dims[T, M]]:
        return self._numpy_array

    @property
    def horizon(self) -> T:
        return cast(T, self._array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self._array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, M]]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxNormals[T: int, M: int](Normals[T, M]):
    _array: Float[JaxArray, "T 2 M"]

    @staticmethod
    def create[T_: int, M_: int](
        *,
        x: Float[JaxArray, "T M"] | Array[Dims[T_, M_]],
        y: Float[JaxArray, "T M"] | Array[Dims[T_, M_]],
        horizon: T_ | None = None,
        rollout_count: M_ | None = None,
    ) -> "JaxNormals[T_, M_]":
        """Creates a JAX normals instance from x and y coordinate arrays."""
        x = jnp.asarray(x)
        y = jnp.asarray(y)
        horizon = horizon if horizon is not None else cast(T_, x.shape[0])
        rollout_count = (
            rollout_count if rollout_count is not None else cast(M_, x.shape[1])
        )

        assert x.shape == y.shape == (horizon, rollout_count), (
            f"Expected x and y to have shape {(horizon, rollout_count)}, "
            f"but got {x.shape} and {y.shape}."
        )

        return JaxNormals(_array=jnp.stack([x, y], axis=1))

    def __array__(self) -> Array[Dims[T, D[2], M]]:
        return self._numpy_array

    def x(self) -> Array[Dims[T, M]]:
        return np.asarray(self._array[:, 0])

    def y(self) -> Array[Dims[T, M]]:
        return np.asarray(self._array[:, 1])

    @property
    def horizon(self) -> T:
        return cast(T, self._array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self._array.shape[2])

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Array[Dims[T, D[2], M]]:
        return np.asarray(self.array)



from typing import Any
from dataclasses import dataclass
from functools import cached_property

from trajax.types import StateSequence, ContouringCost, LagCost, SimulationData, Metric
from trajax.collectors import access

from numtypes import Array, Dims

import numpy as np


@dataclass(kw_only=True, frozen=True)
class MpccErrorMetricResult[T: int = int]:
    contouring: Array[Dims[T]]
    lag: Array[Dims[T]]

    @cached_property
    def max_contouring(self) -> float:
        """Returns the maximum absolute contouring error detected."""
        return float(np.abs(self.contouring).max())

    @cached_property
    def max_lag(self) -> float:
        """Returns the maximum absolute lag error detected."""
        return float(np.abs(self.lag).max())


@dataclass(kw_only=True, frozen=True)
class MpccErrorMetric[StateBatchT](Metric[MpccErrorMetricResult[Any]]):
    contouring: ContouringCost[Any, StateBatchT]
    lag: LagCost[Any, StateBatchT]

    @staticmethod
    def create(
        *,
        contouring: ContouringCost[Any, StateBatchT],
        lag: LagCost[Any, StateBatchT],
    ) -> "MpccErrorMetric[StateBatchT]":
        return MpccErrorMetric(contouring=contouring, lag=lag)

    def compute[T: int = int](self, data: SimulationData) -> MpccErrorMetricResult[T]:
        states = data(
            access.states.assume(StateSequence[T, Any, StateBatchT]).require()
        )
        state_batch = states.batched()

        contouring_error = self.contouring.error(states=state_batch)
        lag_error = self.lag.error(states=state_batch)

        return MpccErrorMetricResult(
            contouring=np.asarray(contouring_error), lag=np.asarray(lag_error)
        )

    @property
    def name(self) -> str:
        return "mpcc-error"



from typing import Protocol

from trajax.types.collectors import SimulationData


class Metric[ResultT](Protocol):
    def compute(self, data: SimulationData) -> ResultT:
        """Computes the metric based on the provided simulation data."""
        ...

    @property
    def name(self) -> str:
        """The name of the metric."""
        ...



from typing import Protocol

from trajax.types.obstacles.history.common import (
    ObstaclePositionsForTimeStep,
    ObstaclePositions,
    ObstaclePositionExtractor,
)

from jaxtyping import Array as JaxArray, Float


class JaxObstaclePositionsForTimeStep[D_p: int, K: int](
    ObstaclePositionsForTimeStep[D_p, K], Protocol
):
    @property
    def array(self) -> Float[JaxArray, "D_p K"]:
        """Returns the obstacle positions for a single time step as a JAX array."""
        ...


class JaxObstaclePositions[T: int, D_p: int, K: int](
    ObstaclePositions[T, D_p, K], Protocol
):
    @property
    def array(self) -> Float[JaxArray, "T D_p K"]:
        """Returns the obstacle positions as a JAX array."""
        ...


class JaxObstaclePositionExtractor[
    ObstacleStatesForTimeStepT,
    ObstacleStatesT,
    PositionsForTimeStepT,
    PositionsT,
](
    ObstaclePositionExtractor[
        ObstacleStatesForTimeStepT, ObstacleStatesT, PositionsForTimeStepT, PositionsT
    ],
    Protocol,
): ...



from .bicycle import (
    BICYCLE_D_X as BICYCLE_D_X,
    BICYCLE_D_V as BICYCLE_D_V,
    BICYCLE_D_U as BICYCLE_D_U,
    BICYCLE_D_O as BICYCLE_D_O,
    BicycleD_x as BicycleD_x,
    BicycleD_v as BicycleD_v,
    BicycleD_u as BicycleD_u,
    BicycleD_o as BicycleD_o,
    BicycleState as BicycleState,
    BicycleStateSequence as BicycleStateSequence,
    BicycleStateBatch as BicycleStateBatch,
    BicyclePositions as BicyclePositions,
    BicycleControlInputSequence as BicycleControlInputSequence,
    BicycleControlInputBatch as BicycleControlInputBatch,
    NumPyBicycleObstacleStatesHistory as NumPyBicycleObstacleStatesHistory,
    JaxBicycleObstacleStatesHistory as JaxBicycleObstacleStatesHistory,
)
from .integrator import (
    IntegratorState as IntegratorState,
    IntegratorStateSequence as IntegratorStateSequence,
    IntegratorStateBatch as IntegratorStateBatch,
    IntegratorControlInputSequence as IntegratorControlInputSequence,
    IntegratorControlInputBatch as IntegratorControlInputBatch,
    NumPyIntegratorState as NumPyIntegratorState,
    NumPyIntegratorStateSequence as NumPyIntegratorStateSequence,
    NumPyIntegratorStateBatch as NumPyIntegratorStateBatch,
    NumPyIntegratorControlInputSequence as NumPyIntegratorControlInputSequence,
    NumPyIntegratorControlInputBatch as NumPyIntegratorControlInputBatch,
    NumPyIntegratorObstacleStatesHistory as NumPyIntegratorObstacleStatesHistory,
    JaxIntegratorState as JaxIntegratorState,
    JaxIntegratorStateSequence as JaxIntegratorStateSequence,
    JaxIntegratorStateBatch as JaxIntegratorStateBatch,
    JaxIntegratorControlInputSequence as JaxIntegratorControlInputSequence,
    JaxIntegratorControlInputBatch as JaxIntegratorControlInputBatch,
    JaxIntegratorObstacleStatesHistory as JaxIntegratorObstacleStatesHistory,
)



from typing import Protocol

from trajax.types.obstacles.common import ObstacleSimulator


class JaxObstacleSimulator[ObstacleStatesForTimeStepT](
    ObstacleSimulator[ObstacleStatesForTimeStepT], Protocol
): ...



from typing import Protocol

from trajax.types.obstacles.history.common import (
    ObstaclePositionsForTimeStep,
    ObstaclePositions,
    ObstaclePositionExtractor,
)

from numtypes import Array, Dims


class NumPyObstaclePositionsForTimeStep[D_p: int, K: int](
    ObstaclePositionsForTimeStep[D_p, K], Protocol
):
    @property
    def array(self) -> Array[Dims[D_p, K]]:
        """Returns the obstacle positions for a single time step as a NumPy array."""
        ...


class NumPyObstaclePositions[T: int, D_p: int, K: int](
    ObstaclePositions[T, D_p, K], Protocol
):
    @property
    def array(self) -> Array[Dims[T, D_p, K]]:
        """Returns the obstacle positions as a NumPy array."""
        ...


class NumPyObstaclePositionExtractor[
    ObstacleStatesForTimeStepT,
    ObstacleStatesT,
    PositionsForTimeStepT,
    PositionsT,
](
    ObstaclePositionExtractor[
        ObstacleStatesForTimeStepT, ObstacleStatesT, PositionsForTimeStepT, PositionsT
    ],
    Protocol,
): ...



from typing import Self, cast
from dataclasses import dataclass

from trajax.types import NumPyObstacleSimulator
from trajax.obstacles.basic import NumPyObstacleStatesForTimeStep

from numtypes import Array, Dims, D, shape_of

import numpy as np


@dataclass(frozen=True)
class NumPyStaticObstacleSimulator[K: int](
    NumPyObstacleSimulator[NumPyObstacleStatesForTimeStep[K]]
):
    positions: Array[Dims[K, D[2]]]
    headings: Array[Dims[K]]

    @staticmethod
    def empty() -> "NumPyStaticObstacleSimulator[D[0]]":
        positions = np.empty((0, 2))

        assert shape_of(positions, matches=(0, 2))

        return NumPyStaticObstacleSimulator.create(positions=positions)

    @staticmethod
    def create[K_: int](
        *, positions: Array[Dims[K_, D[2]]], headings: Array[Dims[K_]] | None = None
    ) -> "NumPyStaticObstacleSimulator[K_]":
        count = positions.shape[0]
        headings = (
            headings
            if headings is not None
            else cast(Array[Dims[K_]], np.zeros(shape=(count,)))
        )

        return NumPyStaticObstacleSimulator(positions=positions, headings=headings)

    def with_time_step_size(self, time_step_size: float) -> Self:
        # Time step does not matter.
        return self

    def step(self) -> NumPyObstacleStatesForTimeStep[K]:
        return NumPyObstacleStatesForTimeStep.create(
            x=self.positions[:, 0],
            y=self.positions[:, 1],
            heading=self.headings,
        )

    @property
    def obstacle_count(self) -> int:
        return self.positions.shape[0]



from dataclasses import dataclass
from functools import cached_property

from trajax.types import (
    jaxtyped,
    D_R,
    Trajectory,
    NumPyPathParameters,
    JaxPathParameters,
    JaxReferencePoints,
    JaxPositions,
    JaxLateralPositions,
    JaxLongitudinalPositions,
    JaxNormals,
)

from jaxtyping import Array as JaxArray, Float, Scalar

import jax
import jax.numpy as jnp


type Vector = Float[JaxArray, "2"]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxLineTrajectory(
    Trajectory[
        JaxPathParameters | NumPyPathParameters,
        JaxReferencePoints,
        JaxPositions,
        JaxLateralPositions,
        JaxLongitudinalPositions,
    ]
):
    start: Vector
    direction: Vector

    heading: Scalar

    _end: tuple[float, float]
    _path_length: Scalar

    @staticmethod
    def create(
        *, start: tuple[float, float], end: tuple[float, float], path_length: float
    ) -> "JaxLineTrajectory":
        """Generates a straight line trajectory from start to end."""
        return JaxLineTrajectory(
            start=(start_array := jnp.array(start)),
            direction=(direction := jnp.array(end) - start_array),
            heading=jnp.arctan2(direction[1], direction[0]),
            _end=end,
            _path_length=jnp.array(path_length),
        )

    def query[T: int, M: int](
        self, parameters: JaxPathParameters[T, M] | NumPyPathParameters[T, M]
    ) -> JaxReferencePoints[T, M]:
        return JaxReferencePoints(
            query(
                parameters=jnp.asarray(parameters.array),
                start=self.start,
                direction=self.direction,
                path_length=self._path_length,
                heading=self.heading,
            )
        )

    def lateral[T: int, M: int](
        self, positions: JaxPositions[T, M]
    ) -> JaxLateralPositions[T, M]:
        return JaxLateralPositions(
            lateral(
                positions=positions.array,
                start=self.start,
                perpendicular=self.perpendicular,
            )
        )

    def longitudinal[T: int, M: int](
        self, positions: JaxPositions[T, M]
    ) -> JaxLongitudinalPositions[T, M]:
        return JaxLongitudinalPositions(
            longitudinal(
                positions=positions.array,
                start=self.start,
                tangent=self.tangent,
                path_length=self._path_length,
                line_length=self.line_length,
            )
        )

    def normal[T: int, M: int](
        self, parameters: JaxPathParameters[T, M] | NumPyPathParameters[T, M]
    ) -> JaxNormals[T, M]:
        T, M = parameters.horizon, parameters.rollout_count

        x = jnp.full((T, M), self.perpendicular[0])
        y = jnp.full((T, M), self.perpendicular[1])

        return JaxNormals.create(x=x, y=y)

    @property
    def end(self) -> tuple[float, float]:
        return self._end

    @property
    def path_length(self) -> float:
        return self._path_length_float

    @property
    def natural_length(self) -> float:
        return self._line_length_float

    @cached_property
    def perpendicular(self) -> Vector:
        tangent = self.tangent
        perpendicular = jnp.array([tangent[1], -tangent[0]])
        return perpendicular

    @cached_property
    def tangent(self) -> Vector:
        return self.direction / self.line_length

    @cached_property
    def line_length(self) -> Scalar:
        return jnp.linalg.norm(self.direction)

    @cached_property
    def _path_length_float(self) -> float:
        return float(self._path_length)

    @cached_property
    def _line_length_float(self) -> float:
        return float(self.line_length)


@jax.jit
@jaxtyped
def query(
    parameters: Float[JaxArray, "T M"],
    *,
    start: Vector,
    direction: Vector,
    path_length: Scalar,
    heading: Scalar,
) -> Float[JaxArray, f"T {D_R} M"]:
    normalized = parameters / path_length
    x, y = (
        start[:, jnp.newaxis, jnp.newaxis]
        + direction[:, jnp.newaxis, jnp.newaxis] * normalized
    )
    heading = jnp.full_like(x, heading)

    return jnp.stack([x, y, heading], axis=1)


@jax.jit
@jaxtyped
def lateral(
    positions: Float[JaxArray, "T 2 M"],
    *,
    start: Vector,
    perpendicular: Vector,
) -> Float[JaxArray, "T M"]:
    relative = positions - start[:, jnp.newaxis]
    return jnp.einsum("tpm,p->tm", relative, perpendicular)


@jax.jit
@jaxtyped
def longitudinal(
    positions: Float[JaxArray, "T 2 M"],
    *,
    start: Vector,
    tangent: Vector,
    path_length: Scalar,
    line_length: Scalar,
) -> Float[JaxArray, "T M"]:
    relative = positions - start[:, jnp.newaxis]
    projection = jnp.einsum("tpm,p->tm", relative, tangent)
    return projection * path_length / line_length



from dataclasses import dataclass

from trajax.types import CostFunction, Costs, CostSumFunction


@dataclass(frozen=True)
class CombinedCost[InputT, StateT, CostsT: Costs]:
    costs: list[CostFunction[InputT, StateT, CostsT]]
    sum: CostSumFunction[CostsT]

    def __post_init__(self) -> None:
        assert len(self.costs) > 0, "At least one cost function must be provided."

    def __call__(self, *, inputs: InputT, states: StateT) -> CostsT:
        costs = [
            cost_function(inputs=inputs, states=states) for cost_function in self.costs
        ]

        return self.sum(costs=costs, initial=costs[0].zero())



from typing import Sequence, Self, NamedTuple, cast
from dataclasses import dataclass
from functools import cached_property

from trajax.types import DataType
from trajax.obstacles.basic import NumPyObstacleStatesForTimeStep, NumPyObstacleStates

from numtypes import NumberArray, IndexArray, Dims, D, shape_of

import numpy as np


@dataclass(frozen=True)
class NumPyObstacleIds[K: int]:
    _array: IndexArray[Dims[K]]

    @staticmethod
    def empty() -> "NumPyObstacleIds[D[0]]":
        return NumPyObstacleIds(np.array([], dtype=np.intp))

    @staticmethod
    def create[K_: int](*, ids: NumberArray[Dims[K_]]) -> "NumPyObstacleIds[K_]":
        return NumPyObstacleIds(ids.astype(np.intp))

    def __array__(self, dtype: DataType | None = None) -> IndexArray[Dims[K]]:
        return self.array

    @property
    def count(self) -> K:
        return self.array.shape[0]

    @property
    def array(self) -> IndexArray[Dims[K]]:
        return self._array


@dataclass(kw_only=True, frozen=True)
class NumPyObstacleStatesRunningHistory[H: int, K: int]:
    class Entry(NamedTuple):
        states: NumPyObstacleStatesForTimeStep
        ids: NumPyObstacleIds | None

    class MergedIds[K_: int](NamedTuple):
        all: NumPyObstacleIds
        recent: NumPyObstacleIds[K_]

    history: list[Entry]
    fixed_horizon: H | None
    fixed_obstacle_count: K | None

    @staticmethod
    def empty[H_: int = int, K_: int = int](
        *, horizon: H_ | None = None, obstacle_count: K_ | None = None
    ) -> "NumPyObstacleStatesRunningHistory[H_, K_]":
        return NumPyObstacleStatesRunningHistory(
            history=[], fixed_horizon=horizon, fixed_obstacle_count=obstacle_count
        )

    @staticmethod
    def single[H_: int = int, K_: int = int](
        observation: NumPyObstacleStatesForTimeStep,
        *,
        horizon: H_ | None = None,
        obstacle_count: K_ | None = None,
    ) -> "NumPyObstacleStatesRunningHistory[H_, K_]":
        return NumPyObstacleStatesRunningHistory(
            history=[NumPyObstacleStatesRunningHistory.Entry(observation, None)],
            fixed_horizon=horizon,
            fixed_obstacle_count=obstacle_count,
        )

    def last(self) -> NumPyObstacleStatesForTimeStep[K]:
        assert self.horizon > 0, "Cannot get last state from empty history."

        return self.history[-1].states

    def get(self) -> NumPyObstacleStates[int, K]:
        return (
            self._combined_history
            if self.horizon > 0
            else NumPyObstacleStates.empty(
                horizon=self.fixed_horizon or 0,
                obstacle_count=self.fixed_obstacle_count or 0,
            )
        )

    def ids(self) -> NumPyObstacleIds[K]:
        return (
            ids.recent
            if (ids := self._merged_ids) is not None
            else NumPyObstacleIds.create(ids=np.array([]))
        )

    def append[N: int](
        self,
        observation: NumPyObstacleStatesForTimeStep,
        *,
        ids: NumPyObstacleIds | None = None,
    ) -> Self:
        assert ids is None or ids.count == observation.count, (
            f"The number of IDs ({ids.count}) does not match "
            f"the number of obstacles in the observation ({observation.count})."
        )

        assert (
            self.fixed_obstacle_count is None
            or observation.count <= self.fixed_obstacle_count
        ), (
            f"Cannot append observation with {observation.count} obstacles to history. "
            f"The obstacle count is fixed to {self.fixed_obstacle_count}."
        )

        entries = self.history + [self.Entry(observation, ids)]

        return self.__class__(
            history=entries[-self.fixed_horizon :]
            if self.fixed_horizon is not None
            else entries,
            fixed_horizon=self.fixed_horizon,
            fixed_obstacle_count=self.fixed_obstacle_count,
        )

    @property
    def horizon(self) -> int:
        return len(self.history)

    @property
    def count(self) -> K:
        return self._count

    @cached_property
    def _count(self) -> K:
        if self.fixed_obstacle_count is not None:
            return self.fixed_obstacle_count

        return ids.all.count if (ids := self._merged_ids) is not None else cast(K, 0)

    @cached_property
    def _full_horizon(self) -> int:
        if self.fixed_horizon is not None:
            return self.fixed_horizon

        return len(self.history)

    @cached_property
    def _combined_history(self) -> NumPyObstacleStates[int, K]:
        if (ids := self._merged_ids) is None:
            return NumPyObstacleStates.wrap(
                np.stack([entry.states.array for entry in self.history], axis=0)
            )

        return combine_history(
            recent_ids=ids.recent,
            history=self.history,
            horizon=self._full_horizon,
            obstacle_count=self.count,
        )

    @cached_property
    def _merged_ids(self) -> "NumPyObstacleStatesRunningHistory.MergedIds[K] | None":
        if all(entry.ids is None for entry in self.history):
            return None

        # NOTE: A dict is used here to maintain insertion order while ensuring uniqueness.
        seen_ids: dict[int, None] = {}

        # NOTE: We iterate in reverse to get most recent IDs first.
        for entry in reversed(self.history):
            assert entry.ids is not None, (
                f"Missing IDs in history entry {entry} while others have IDs."
            )

            for obstacle_id in entry.ids.array:
                # NOTE: We do not overwrite existing IDs to maintain most recent occurrence.
                if (typed_id := int(obstacle_id)) not in seen_ids:
                    seen_ids[typed_id] = None

        all_ids = list(seen_ids.keys())
        recent_ids = (
            all_ids[: self.fixed_obstacle_count]
            if self.fixed_obstacle_count is not None
            else all_ids
        )

        return self.MergedIds(
            all=NumPyObstacleIds.create(ids=np.sort(np.array(all_ids, dtype=np.intp))),
            recent=NumPyObstacleIds.create(
                ids=np.sort(np.array(recent_ids, dtype=np.intp))
            ),
        )


def combine_history[H: int, K: int](
    *,
    recent_ids: NumPyObstacleIds[K],
    history: Sequence[NumPyObstacleStatesRunningHistory.Entry],
    horizon: H,
    obstacle_count: K,
) -> NumPyObstacleStates[H, K]:
    recent_id_count = recent_ids.count
    time_padding = horizon - len(history)
    dimension = history[0].states.dimension
    output = np.full((output_shape := (horizon, dimension, obstacle_count)), np.nan)

    for t, entry in enumerate(history):
        assert entry.ids is not None, (
            f"Missing IDs in history entry {entry} while others have IDs."
        )

        positions = np.clip(
            np.searchsorted(recent_ids.array, entry.ids.array), 0, recent_id_count - 1
        )

        valid_mask = (positions < recent_id_count) & (
            recent_ids.array[positions] == entry.ids.array
        )

        valid_positions = positions[valid_mask]
        output[time_padding + t, :, valid_positions] = entry.states.array[
            :, valid_mask
        ].T

    assert shape_of(output, matches=output_shape)

    return NumPyObstacleStates.wrap(output)



from typing import Any
from dataclasses import dataclass
from functools import cached_property

from trajax.types import (
    StateSequence,
    SimulationData,
    Metric,
    Positions,
    LateralPositions,
    PositionExtractor,
    Trajectory,
    BoundaryDistance,
    BoundaryDistanceExtractor,
)
from trajax.collectors import access

from numtypes import Array, BoolArray, Dims

import numpy as np


@dataclass(kw_only=True, frozen=True)
class ConstraintViolationMetricResult[T: int = int]:
    lateral_deviations: Array[Dims[T]]
    boundary_distances: Array[Dims[T]]

    @cached_property
    def violations(self) -> BoolArray[Dims[T]]:
        return self.boundary_distances <= 0

    @cached_property
    def violation_detected(self) -> bool:
        return bool(self.violations.any())


@dataclass(kw_only=True, frozen=True)
class ConstraintViolationMetric[StateBatchT, PositionsT, LateralT, BoundaryDistanceT](
    Metric[ConstraintViolationMetricResult[Any]]
):
    reference: Trajectory[Any, Any, PositionsT, LateralT]
    boundary: BoundaryDistanceExtractor[StateBatchT, BoundaryDistanceT]
    position_extractor: PositionExtractor[StateBatchT, PositionsT]

    @staticmethod
    def create[S, P: Positions, L: LateralPositions, BD: BoundaryDistance](
        *,
        reference: Trajectory[Any, Any, P, L],
        boundary: BoundaryDistanceExtractor[S, BD],
        position_extractor: PositionExtractor[S, P],
    ) -> "ConstraintViolationMetric[S, P, L, BD]":
        return ConstraintViolationMetric(
            reference=reference,
            boundary=boundary,
            position_extractor=position_extractor,
        )

    def compute[T: int = int](
        self, data: SimulationData
    ) -> ConstraintViolationMetricResult[T]:
        states = data(
            access.states.assume(StateSequence[T, Any, StateBatchT]).require()
        )
        state_batch = states.batched()

        positions = self.position_extractor(state_batch)
        lateral = self.reference.lateral(positions)
        boundary_distance = self.boundary(states=state_batch)

        lateral_array = np.asarray(lateral)[:, 0]
        boundary_array = np.asarray(boundary_distance)[:, 0]

        return ConstraintViolationMetricResult(
            lateral_deviations=lateral_array, boundary_distances=boundary_array
        )

    @property
    def name(self) -> str:
        return "constraint-violation"



from typing import Protocol

from trajax.types.array import DataType

from numtypes import Array, Dims


class Error[T: int, M: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        """Returns the error as a NumPy array."""
        ...


class PositionExtractor[StateBatchT, PositionsT](Protocol):
    def __call__(self, states: StateBatchT, /) -> PositionsT:
        """Extracts (x, y) positions from a batch of states."""
        ...


class ContouringCost[InputBatchT, StateBatchT, ErrorT = Error](Protocol):
    def error(self, *, states: StateBatchT) -> ErrorT:
        """Computes the contouring error for the given states."""
        ...


class LagCost[InputBatchT, StateBatchT, ErrorT = Error](Protocol):
    def error(self, *, states: StateBatchT) -> ErrorT:
        """Computes the lag error for the given states."""
        ...



from .history import (
    ObstaclePositionsForTimeStep as ObstaclePositionsForTimeStep,
    ObstaclePositions as ObstaclePositions,
    ObstaclePositionExtractor as ObstaclePositionExtractor,
    NumPyObstaclePositionsForTimeStep as NumPyObstaclePositionsForTimeStep,
    NumPyObstaclePositions as NumPyObstaclePositions,
    NumPyObstaclePositionExtractor as NumPyObstaclePositionExtractor,
    JaxObstaclePositionsForTimeStep as JaxObstaclePositionsForTimeStep,
    JaxObstaclePositions as JaxObstaclePositions,
    JaxObstaclePositionExtractor as JaxObstaclePositionExtractor,
)
from .common import (
    ObstacleSimulator as ObstacleSimulator,
)
from .basic import (
    NumPyObstacleSimulator as NumPyObstacleSimulator,
)
from .accelerated import (
    JaxObstacleSimulator as JaxObstacleSimulator,
)



from .common import (
    EstimatedObstacleStates as EstimatedObstacleStates,
    ObstacleStatesHistory as ObstacleStatesHistory,
    ObstacleIds as ObstacleIds,
    ObstacleIdAssignment as ObstacleIdAssignment,
    ObstacleStatesRunningHistory as ObstacleStatesRunningHistory,
    ObstacleStateSequences as ObstacleStateSequences,
    CovarianceSequences as CovarianceSequences,
    ObstacleModel as ObstacleModel,
    PredictionCreator as PredictionCreator,
    CovariancePropagator as CovariancePropagator,
    ObstacleMotionPredictor as ObstacleMotionPredictor,
)
from .basic import (
    NumPyObstacleStatesHistory as NumPyObstacleStatesHistory,
)
from .accelerated import (
    JaxObstacleStatesHistory as JaxObstacleStatesHistory,
)
from .propagators import (
    NumPyInitialPositionCovariance as NumPyInitialPositionCovariance,
    NumPyInitialVelocityCovariance as NumPyInitialVelocityCovariance,
    NumPyPositionCovariance as NumPyPositionCovariance,
    NumPyInitialCovarianceProvider as NumPyInitialCovarianceProvider,
    JaxInitialPositionCovariance as JaxInitialPositionCovariance,
    JaxInitialVelocityCovariance as JaxInitialVelocityCovariance,
    JaxPositionCovariance as JaxPositionCovariance,
    JaxInitialCovarianceProvider as JaxInitialCovarianceProvider,
)



from .basic import (
    NumPyConstantVarianceProvider as NumPyConstantVarianceProvider,
    NumPyConstantCovarianceProvider as NumPyConstantCovarianceProvider,
)
from .accelerated import (
    JaxConstantVarianceProvider as JaxConstantVarianceProvider,
    JaxConstantCovarianceProvider as JaxConstantCovarianceProvider,
)



from typing import overload, Literal, NamedTuple
from dataclasses import dataclass

from trajax.types import (
    jaxtyped,
    D_R,
    Trajectory,
    NumPyPathParameters,
    JaxPathParameters,
    JaxReferencePoints,
    JaxPositions,
    JaxLateralPositions,
    JaxLongitudinalPositions,
    JaxNormals,
)
from trajax.trajectories.waypoints.basic import NumPyWaypointsTrajectory

from scipy.interpolate import CubicSpline
from jaxtyping import Array as JaxArray, Float, Scalar, Int as JaxInt
from numtypes import Array, Dims, D

import jax
import jax.numpy as jnp
import numpy as np

type PointArray = Array[Dims[int, D[2]]]
type JaxPointArray = Float[JaxArray, "N 2"]
type Int = JaxInt[JaxArray, ""]


class GuessSamples(NamedTuple):
    x: Float[JaxArray, "K"]
    y: Float[JaxArray, "K"]
    arc_lengths: Float[JaxArray, "K"]


@dataclass(kw_only=True, frozen=True)
class JaxWaypointsTrajectory(
    Trajectory[
        JaxPathParameters | NumPyPathParameters,
        JaxReferencePoints,
        JaxPositions,
        JaxLateralPositions,
        JaxLongitudinalPositions,
    ]
):
    reference_points: Float[JaxArray, "N"]
    coefficients_x: Float[JaxArray, "N-1 4"]
    coefficients_y: Float[JaxArray, "N-1 4"]

    guess_samples: GuessSamples
    refining_iterations: int

    _path_length: Scalar
    _inner: NumPyWaypointsTrajectory

    @overload
    @staticmethod
    def create(
        *,
        points: PointArray,
        path_length: float,
        coarse_samples: int = 200,
        refining_iterations: int = 3,
    ) -> "JaxWaypointsTrajectory":
        """Creates a waypoints trajectory from a set of 2D points."""
        ...

    @overload
    @staticmethod
    def create(
        *,
        points: JaxPointArray,
        path_length: float,
        coarse_samples: int = 200,
        refining_iterations: int = 3,
    ) -> "JaxWaypointsTrajectory":
        """Creates a waypoints trajectory from a set of 2D points."""
        ...

    @staticmethod
    def create(
        *,
        points: PointArray | JaxPointArray,
        path_length: float,
        coarse_samples: int = 200,
        refining_iterations: int = 3,
    ) -> "JaxWaypointsTrajectory":
        """Creates a waypoints trajectory from a set of 2D points."""
        trajectory = NumPyWaypointsTrajectory.create(
            points=np.asarray(points), path_length=path_length
        )

        return JaxWaypointsTrajectory(
            reference_points=jnp.asarray(trajectory.reference_points),
            coefficients_x=coefficients_from(trajectory.spline_x),
            coefficients_y=coefficients_from(trajectory.spline_y),
            guess_samples=compute_guess_samples(
                trajectory=trajectory, sample_count=coarse_samples
            ),
            refining_iterations=refining_iterations,
            _path_length=jnp.asarray(path_length),
            _inner=trajectory,
        )

    def query[T: int, M: int](
        self, parameters: JaxPathParameters[T, M] | NumPyPathParameters[T, M]
    ) -> JaxReferencePoints[T, M]:
        parameters_array = jnp.asarray(parameters.array)

        assert path_parameters_are_valid(
            parameters_array, path_length=self._path_length
        )

        return JaxReferencePoints(
            query(
                parameters=parameters_array,
                reference_points=self.reference_points,
                coefficients_x=self.coefficients_x,
                coefficients_y=self.coefficients_y,
            )
        )

    def lateral[T: int, M: int](
        self, positions: JaxPositions[T, M]
    ) -> JaxLateralPositions[T, M]:
        return JaxLateralPositions(
            compute_lateral(
                x=positions.x_array,
                y=positions.y_array,
                arc_lengths=closest_arc_lengths(
                    x=positions.x_array,
                    y=positions.y_array,
                    reference_points=self.reference_points,
                    coefficients_x=self.coefficients_x,
                    coefficients_y=self.coefficients_y,
                    guess=self.guess_samples,
                    path_length=self._path_length,
                    refining_iterations=self.refining_iterations,
                ),
                reference_points=self.reference_points,
                coefficients_x=self.coefficients_x,
                coefficients_y=self.coefficients_y,
            )
        )

    def longitudinal[T: int, M: int](
        self, positions: JaxPositions[T, M]
    ) -> JaxLongitudinalPositions[T, M]:
        return JaxLongitudinalPositions(
            closest_arc_lengths(
                x=positions.x_array,
                y=positions.y_array,
                reference_points=self.reference_points,
                coefficients_x=self.coefficients_x,
                coefficients_y=self.coefficients_y,
                guess=self.guess_samples,
                path_length=self._path_length,
                refining_iterations=self.refining_iterations,
            )
        )

    def normal[T: int, M: int](
        self, parameters: JaxPathParameters[T, M] | NumPyPathParameters[T, M]
    ) -> JaxNormals[T, M]:
        x, y = compute_normal(
            parameters=jnp.asarray(parameters.array),
            reference_points=self.reference_points,
            coefficients_x=self.coefficients_x,
            coefficients_y=self.coefficients_y,
        )

        return JaxNormals.create(x=x, y=y)

    @property
    def end(self) -> tuple[float, float]:
        return self._inner.end

    @property
    def path_length(self) -> float:
        return self._inner.path_length

    @property
    def natural_length(self) -> float:
        return self._inner.natural_length


def report_invalid_parameters(valid: bool, parameters: Float[JaxArray, "S 4"]) -> None:
    if not valid:
        print(f"Path parameters out of bounds. Got: {parameters}")


def coefficients_from(spline: CubicSpline) -> Float[JaxArray, "S 4"]:
    # Extracts cubic spline coefficients from scipy CubicSpline.
    #
    # Returns coefficients [a, b, c, d] for each segment where:
    # p(t) = a + b*t + c*t^2 + d*t^3
    # with t being the local parameter within each segment.

    # SciPy stores coefficients in shape (n_segments, 4) with columns [d, c, b, a]
    # We need to reorder to [a, b, c, d]
    # Transpose to get (n_segments, 4), Reverse to get [a, b, c, d]
    return jnp.asarray(spline.c.T[:, ::-1])


def compute_guess_samples(
    trajectory: NumPyWaypointsTrajectory, *, sample_count: int
) -> GuessSamples:
    sample_s = np.linspace(0, trajectory.path_length, sample_count)
    sample_x = trajectory.spline_x(sample_s)
    sample_y = trajectory.spline_y(sample_s)

    return GuessSamples(
        x=jnp.asarray(sample_x),
        y=jnp.asarray(sample_y),
        arc_lengths=jnp.asarray(sample_s),
    )


@jax.jit
@jaxtyped
def path_parameters_are_valid(
    path_parameters: Float[JaxArray, "T M"], *, path_length: Scalar
) -> Literal[True]:
    valid = jnp.all((0.0 <= path_parameters) & (path_parameters <= path_length))
    jax.debug.callback(report_invalid_parameters, valid, path_parameters)
    return True


@jax.jit
@jaxtyped
def query(
    parameters: Float[JaxArray, "T M"],
    *,
    reference_points: Float[JaxArray, "N"],
    coefficients_x: Float[JaxArray, "S 4"],
    coefficients_y: Float[JaxArray, "S 4"],
) -> Float[JaxArray, f"T {D_R} M"]:
    x = evaluate(parameters, reference_points, coefficients_x)
    y = evaluate(parameters, reference_points, coefficients_y)

    dx_ds = evaluate_derivative(parameters, reference_points, coefficients_x)
    dy_ds = evaluate_derivative(parameters, reference_points, coefficients_y)
    heading = jnp.arctan2(dy_ds, dx_ds)

    return jnp.stack([x, y, heading], axis=1)


@jax.jit
@jaxtyped
def evaluate(
    parameters: Float[JaxArray, "T M"],
    reference_points: Float[JaxArray, "N"],
    coefficients: Float[JaxArray, "S 4"],
) -> Float[JaxArray, "T M"]:
    num_segments = len(coefficients)
    segment_indices = jnp.clip(
        jnp.searchsorted(reference_points[1:], parameters, side="right"),
        0,
        num_segments - 1,
    )
    segment_starts = reference_points[segment_indices]
    t = parameters - segment_starts

    segment_coefficients = coefficients[segment_indices]
    a, b, c, d = (
        segment_coefficients[..., 0],
        segment_coefficients[..., 1],
        segment_coefficients[..., 2],
        segment_coefficients[..., 3],
    )

    return a + b * t + c * t**2 + d * t**3


@jax.jit
@jaxtyped
def evaluate_derivative(
    parameters: Float[JaxArray, "T M"],
    reference_points: Float[JaxArray, "N"],
    coefficients: Float[JaxArray, "S 4"],
) -> Float[JaxArray, "T M"]:
    num_segments = len(coefficients)
    segment_indices = jnp.clip(
        jnp.searchsorted(reference_points[1:], parameters, side="right"),
        0,
        num_segments - 1,
    )
    segment_starts = reference_points[segment_indices]
    t = parameters - segment_starts

    segment_coefficients = coefficients[segment_indices]
    b = segment_coefficients[..., 1]
    c = segment_coefficients[..., 2]
    d = segment_coefficients[..., 3]

    return b + 2 * c * t + 3 * d * t**2


@jax.jit
@jaxtyped
def evaluate_second_derivative(
    parameters: Float[JaxArray, "T M"],
    reference_points: Float[JaxArray, "N"],
    coefficients: Float[JaxArray, "S 4"],
) -> Float[JaxArray, "T M"]:
    num_segments = len(coefficients)
    segment_indices = jnp.clip(
        jnp.searchsorted(reference_points[1:], parameters, side="right"),
        0,
        num_segments - 1,
    )
    segment_starts = reference_points[segment_indices]
    t = parameters - segment_starts

    segment_coefficients = coefficients[segment_indices]
    c = segment_coefficients[..., 2]
    d = segment_coefficients[..., 3]

    return 2 * c + 6 * d * t


@jax.jit(static_argnames=("refining_iterations",))
@jaxtyped
def closest_arc_lengths(
    *,
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    reference_points: Float[JaxArray, "N"],
    coefficients_x: Float[JaxArray, "S 4"],
    coefficients_y: Float[JaxArray, "S 4"],
    guess: GuessSamples,
    path_length: Scalar,
    refining_iterations: int,
) -> Float[JaxArray, "T M"]:
    return refine_newton(
        arc_lengths=nearest_sample(x=x, y=y, guess=guess),
        x=x,
        y=y,
        reference_points=reference_points,
        coefficients_x=coefficients_x,
        coefficients_y=coefficients_y,
        path_length=path_length,
        iterations=refining_iterations,
    )


@jax.jit
@jaxtyped
def nearest_sample(
    *,
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    guess: GuessSamples,
) -> Float[JaxArray, "T M"]:
    # Find nearest pre-computed sample for each position (coarse search).
    diff_x = guess.x[:, None, None] - x[None, :, :]
    diff_y = guess.y[:, None, None] - y[None, :, :]
    distances_sq = diff_x**2 + diff_y**2

    nearest_idx = jnp.argmin(distances_sq, axis=0)
    return guess.arc_lengths[nearest_idx]


@jax.jit(static_argnames=("iterations",))
@jaxtyped
def refine_newton(
    *,
    arc_lengths: Float[JaxArray, "T M"],
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    reference_points: Float[JaxArray, "N"],
    coefficients_x: Float[JaxArray, "S 4"],
    coefficients_y: Float[JaxArray, "S 4"],
    path_length: Scalar,
    iterations: int,
) -> Float[JaxArray, "T M"]:
    # Refine arc length estimates using Newton's method.

    def newton_step(
        s: Float[JaxArray, "T M"], _: None
    ) -> tuple[Float[JaxArray, "T M"], None]:
        cx = evaluate(s, reference_points, coefficients_x)
        cy = evaluate(s, reference_points, coefficients_y)

        dx = evaluate_derivative(s, reference_points, coefficients_x)
        dy = evaluate_derivative(s, reference_points, coefficients_y)

        ddx = evaluate_second_derivative(s, reference_points, coefficients_x)
        ddy = evaluate_second_derivative(s, reference_points, coefficients_y)

        # f(s) = (p - c(s)) · c'(s) = 0  (orthogonality condition)
        f = (x - cx) * dx + (y - cy) * dy

        # f'(s) = -|c'(s)|² + (p - c(s)) · c''(s)
        f_prime = -(dx**2 + dy**2) + (x - cx) * ddx + (y - cy) * ddy

        # Newton step
        s_new = s - f / f_prime
        s_new = jnp.clip(s_new, 0.0, path_length)

        return s_new, None

    refined, _ = jax.lax.scan(newton_step, arc_lengths, None, length=iterations)
    return refined


@jax.jit
@jaxtyped
def compute_lateral(
    *,
    x: Float[JaxArray, "T M"],
    y: Float[JaxArray, "T M"],
    arc_lengths: Float[JaxArray, "T M"],
    reference_points: Float[JaxArray, "N"],
    coefficients_x: Float[JaxArray, "S 4"],
    coefficients_y: Float[JaxArray, "S 4"],
) -> Float[JaxArray, "T M"]:
    # Compute signed lateral deviation from trajectory.
    closest_x = evaluate(arc_lengths, reference_points, coefficients_x)
    closest_y = evaluate(arc_lengths, reference_points, coefficients_y)

    dx_ds = evaluate_derivative(arc_lengths, reference_points, coefficients_x)
    dy_ds = evaluate_derivative(arc_lengths, reference_points, coefficients_y)
    tangent_norm = jnp.sqrt(dx_ds**2 + dy_ds**2)

    diff_x = x - closest_x
    diff_y = y - closest_y

    # NOTE: Cross product for signed lateral (right = positive, left = negative)
    lateral = (diff_x * dy_ds - diff_y * dx_ds) / tangent_norm

    return lateral


@jax.jit
@jaxtyped
def compute_normal(
    parameters: Float[JaxArray, "T M"],
    *,
    reference_points: Float[JaxArray, "N"],
    coefficients_x: Float[JaxArray, "S 4"],
    coefficients_y: Float[JaxArray, "S 4"],
) -> tuple[Float[JaxArray, "T M"], Float[JaxArray, "T M"]]:
    dx_ds = evaluate_derivative(parameters, reference_points, coefficients_x)
    dy_ds = evaluate_derivative(parameters, reference_points, coefficients_y)
    norm = jnp.sqrt(dx_ds**2 + dy_ds**2)

    normal_x = dy_ds / norm
    normal_y = -dx_ds / norm

    return normal_x, normal_y



from typing import Any
from dataclasses import dataclass

from trajax.types import (
    ObstacleStateSequences,
    NumPyInitialPositionCovariance,
    NumPyInitialVelocityCovariance,
)

import numpy as np


@dataclass(kw_only=True, frozen=True)
class NumPyConstantVarianceProvider:
    position_variance: float
    velocity_variance: float

    @staticmethod
    def create(
        *,
        position_variance: float,
        velocity_variance: float,
    ) -> "NumPyConstantVarianceProvider":
        return NumPyConstantVarianceProvider(
            position_variance=position_variance, velocity_variance=velocity_variance
        )

    def position[K: int](
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> NumPyInitialPositionCovariance[K]:
        return np.tile(
            (np.eye(2) * self.position_variance)[..., np.newaxis], (1, 1, states.count)
        )

    def velocity[K: int](
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> NumPyInitialVelocityCovariance[K]:
        return np.tile(
            (np.eye(2) * self.velocity_variance)[..., np.newaxis], (1, 1, states.count)
        )


@dataclass(kw_only=True, frozen=True)
class NumPyConstantCovarianceProvider[K: int]:
    position_covariance: NumPyInitialPositionCovariance[K]
    velocity_covariance: NumPyInitialVelocityCovariance[K]

    @staticmethod
    def create(
        *,
        position_covariance: NumPyInitialPositionCovariance[K],
        velocity_covariance: NumPyInitialVelocityCovariance[K],
    ) -> "NumPyConstantCovarianceProvider[K]":
        return NumPyConstantCovarianceProvider(
            position_covariance=position_covariance,
            velocity_covariance=velocity_covariance,
        )

    def position(
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> NumPyInitialPositionCovariance[K]:
        assert states.count == self.position_covariance.shape[2]
        return self.position_covariance

    def velocity(
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> NumPyInitialVelocityCovariance[K]:
        assert states.count == self.velocity_covariance.shape[2]
        return self.velocity_covariance



from typing import NamedTuple


class MpccMppiSetup[MppiT, ModelT, ContouringCostT, LagCostT](NamedTuple):
    mppi: MppiT
    model: ModelT
    contouring_cost: ContouringCostT
    lag_cost: LagCostT
